from __future__ import annotations

import pytest
import stk

from ...case_data import CaseData


def _get_cage() -> stk.ConstructedMolecule:
    return stk.ConstructedMolecule(
        topology_graph=stk.cage.FourPlusSix(
            building_blocks={
                stk.BuildingBlock(
                    smiles="BrC1=C(Br)[C+]=[C+]1",
                    functional_groups=[stk.BromoFactory()],
                ): (4, 5, 6, 7, 8),
                stk.BuildingBlock(
                    smiles="BrC1=C(Br)[C+]=N1",
                    functional_groups=[stk.BromoFactory()],
                ): (9,),
                stk.BuildingBlock(
                    smiles=("Br[C+]1[C+2][C+](Br)[C+]2[C+][C+2]C2(Br)[C+2]1"),
                    functional_groups=[stk.BromoFactory()],
                ): (0, 1, 2),
                stk.BuildingBlock(
                    smiles=(
                        "Br[C+]1[C+2][C+](Br)[C+]2[C+](F)[C+2]C2(Br)[C+2]1"
                    ),
                    functional_groups=[stk.BromoFactory()],
                ): (3,),
            },
        ),
    )


def _get_guests() -> tuple[stk.host_guest.Guest, ...]:
    return (
        stk.host_guest.Guest(stk.BuildingBlock("C#N")),
        stk.host_guest.Guest(stk.BuildingBlock("C#C")),
    )


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.host_guest.Complex(
                    host=stk.BuildingBlock.init_from_molecule(
                        molecule=_get_cage(),
                    ),
                    guests=stk.host_guest.Guest(
                        building_block=stk.BuildingBlock("C#N")
                    ),
                )
            ),
            smiles=(
                "F[C+]1[C+2]C23[C+2][C+]4[C+2][C+](C5=C(N=[C+]5)[C+]5"
                "[C+2][C+]6[C+2]C7([C+2][CH+][C+]57)C5=C([C+]=[C+]5)["
                "C+]5[C+2][C+]([C+2]C7([C+2][CH+][C+]57)C5=C2[C+]=[C+"
                "]5)C2=C([C+]=[C+]2)[C+]2[C+2][C+](C5=C4[C+]=[C+]5)[C"
                "+]4[CH+][C+2]C4([C+2]2)C2=C6[C+]=[C+]2)[C+]13.[H]C#N"
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.host_guest.Complex(
                    host=stk.BuildingBlock.init_from_molecule(
                        molecule=_get_cage(),
                    ),
                    guests=_get_guests(),
                )
            ),
            smiles=(
                "F[C+]1[C+2]C23[C+2][C+]4[C+2][C+](C5=C(N=[C+]5)[C+]5"
                "[C+2][C+]6[C+2]C7([C+2][CH+][C+]57)C5=C([C+]=[C+]5)["
                "C+]5[C+2][C+]([C+2]C7([C+2][CH+][C+]57)C5=C2[C+]=[C+"
                "]5)C2=C([C+]=[C+]2)[C+]2[C+2][C+](C5=C4[C+]=[C+]5)[C"
                "+]4[CH+][C+2]C4([C+2]2)C2=C6[C+]=[C+]2)[C+]13"
                ".[H]C#C[H].[H]C#N"
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.host_guest.Complex(
                    host=stk.BuildingBlock.init_from_molecule(
                        molecule=_get_cage(),
                    ),
                    guests=(i for i in _get_guests()),
                )
            ),
            smiles=(
                "F[C+]1[C+2]C23[C+2][C+]4[C+2][C+](C5=C(N=[C+]5)[C+]5"
                "[C+2][C+]6[C+2]C7([C+2][CH+][C+]57)C5=C([C+]=[C+]5)["
                "C+]5[C+2][C+]([C+2]C7([C+2][CH+][C+]57)C5=C2[C+]=[C+"
                "]5)C2=C([C+]=[C+]2)[C+]2[C+2][C+](C5=C4[C+]=[C+]5)[C"
                "+]4[CH+][C+2]C4([C+2]2)C2=C6[C+]=[C+]2)[C+]13"
                ".[H]C#C[H].[H]C#N"
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.host_guest.Complex(
                    host=stk.BuildingBlock.init_from_molecule(
                        molecule=_get_cage(),
                    ),
                    guests=(i for i in _get_guests()),
                    optimizer=stk.Spinner(),
                )
            ),
            smiles=(
                "F[C+]1[C+2]C23[C+2][C+]4[C+2][C+](C5=C(N=[C+]5)[C+]5"
                "[C+2][C+]6[C+2]C7([C+2][CH+][C+]57)C5=C([C+]=[C+]5)["
                "C+]5[C+2][C+]([C+2]C7([C+2][CH+][C+]57)C5=C2[C+]=[C+"
                "]5)C2=C([C+]=[C+]2)[C+]2[C+2][C+](C5=C4[C+]=[C+]5)[C"
                "+]4[CH+][C+2]C4([C+2]2)C2=C6[C+]=[C+]2)[C+]13"
                ".[H]C#C[H].[H]C#N"
            ),
            name=name,
        ),
    ),
)
def host_guest_complex(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



from stk._internal.construction_state.construction_state import (
    ConstructionState,
)


class Optimizer:
    """
    An abstract base class for optimizers.

    An optimizer is used to change the structure of the molecule under
    construction to be more realistic.

    """

    def optimize(self, state: ConstructionState) -> ConstructionState:
        """
        Optimize the structure of a molecule under construction.

        Parameters:
            state:
                The molecule being constructed.

        Returns:
            The optimized construction state.

        """

        raise NotImplementedError()



import pytest
import stk
from collections.abc import Sequence
from ..case_data import CaseData
from dataclasses import dataclass


@pytest.fixture(
    params=["building_block.mol"],
)
def path(tmpdir, request):
    return tmpdir / request.param


@dataclass(frozen=True, slots=True)
class InitFromFileData:
    building_block: stk.BuildingBlock
    init_functional_groups: Sequence[stk.FunctionalGroup]
    init_placer_ids: Sequence[int] | None
    case_data_functional_groups: Sequence[stk.FunctionalGroup]
    case_data_core_atom_ids: Sequence[int]
    case_data_placer_ids: Sequence[int]
    known_repr: str


@pytest.fixture(
    scope="session",
    params=(
        lambda: InitFromFileData(
            building_block=stk.BuildingBlock("Br[C+2][C+2]Br"),
            init_functional_groups=(),
            init_placer_ids=None,
            case_data_functional_groups=(),
            case_data_core_atom_ids=(0, 1, 2, 3),
            case_data_placer_ids=(0, 1, 2, 3),
            known_repr="BuildingBlock('Br[C+2][C+2]Br', ())",
        ),
        lambda: InitFromFileData(
            building_block=stk.BuildingBlock("Br[C+2][C+2]Br"),
            init_functional_groups=[stk.BromoFactory()],
            init_placer_ids=None,
            case_data_functional_groups=(
                stk.Bromo(
                    bromine=stk.Br(0),
                    atom=stk.C(1, 2),
                    bonders=(stk.C(1, 2),),
                    deleters=(stk.Br(0),),
                ),
                stk.Bromo(
                    bromine=stk.Br(3),
                    atom=stk.C(2, 2),
                    bonders=(stk.C(2, 2),),
                    deleters=(stk.Br(3),),
                ),
            ),
            case_data_core_atom_ids=(1, 2),
            case_data_placer_ids=(1, 2),
            known_repr=(
                "BuildingBlock('Br[C+2][C+2]Br', (Bromo(Br(0), C(1, charge=2),"
                " bonders=(C(1, charge=2),), deleters=(Br(0),)), Bromo(Br(3), "
                "C(2, charge=2), bonders=(C(2, charge=2),), deleters=(Br(3),))"
                "))"
            ),
        ),
        lambda: InitFromFileData(
            building_block=stk.BuildingBlock("Br[C+2][C+2]Br"),
            init_functional_groups=(),
            init_placer_ids=(1, 2),
            case_data_functional_groups=(),
            case_data_core_atom_ids=(0, 1, 2, 3),
            case_data_placer_ids=(1, 2),
            known_repr="BuildingBlock('Br[C+2][C+2]Br', ())",
        ),
        lambda: InitFromFileData(
            building_block=stk.BuildingBlock("Br[C+2][C+2]Br"),
            init_functional_groups=[stk.BromoFactory()],
            init_placer_ids=(0, 3),
            case_data_functional_groups=(
                stk.Bromo(
                    bromine=stk.Br(0),
                    atom=stk.C(1, 2),
                    bonders=(stk.C(1, 2),),
                    deleters=(stk.Br(0),),
                ),
                stk.Bromo(
                    bromine=stk.Br(3),
                    atom=stk.C(2, 2),
                    bonders=(stk.C(2, 2),),
                    deleters=(stk.Br(3),),
                ),
            ),
            case_data_core_atom_ids=(1, 2),
            case_data_placer_ids=(0, 3),
            known_repr=(
                "BuildingBlock('Br[C+2][C+2]Br', (Bromo(Br(0), C(1, charge=2),"
                " bonders=(C(1, charge=2),), deleters=(Br(0),)), Bromo(Br(3), "
                "C(2, charge=2), bonders=(C(2, charge=2),), deleters=(Br(3),))"
                "))"
            ),
        ),
        lambda: InitFromFileData(
            building_block=stk.BuildingBlock("Br[C+2][C+2]Br"),
            init_functional_groups=[stk.IodoFactory()],
            init_placer_ids=None,
            case_data_functional_groups=(),
            case_data_core_atom_ids=(0, 1, 2, 3),
            case_data_placer_ids=(0, 1, 2, 3),
            known_repr="BuildingBlock('Br[C+2][C+2]Br', ())",
        ),
    ),
)
def init_from_file_data(request) -> InitFromFileData:
    """
    A :class:`.InitFromFileData` instance.

    """

    return request.param()


@pytest.fixture
def init_from_file(path, init_from_file_data):
    """
    A :class:`.CaseData` instance.

    """

    data = init_from_file_data
    data.building_block.write(path)
    return CaseData(
        building_block=stk.BuildingBlock.init_from_file(
            path=path,
            functional_groups=data.init_functional_groups,
            placer_ids=data.init_placer_ids,
        ),
        functional_groups=data.case_data_functional_groups,
        core_atom_ids=data.case_data_core_atom_ids,
        placer_ids=data.case_data_placer_ids,
        known_repr=data.known_repr,
    )



import numpy as np

from .utilities import is_clone


def test_with_position(vertex, position):
    """
    Test :meth:`.Vertex.with_position`.

    Parameters
    ----------
    vertex : :class:`.Vertex`
        The vertex to test.

    position : :class:`numpy.ndarray`
        The position of the new vertex.

    Returns
    -------
    None : :class:`NoneType`

    """

    # Save a clone to check immutability.
    clone = vertex.clone()
    _test_with_position(vertex, position)
    is_clone(vertex, clone)


def _test_with_position(vertex, position):
    """
    Test :meth:`.Vertex.with_position`.

    Parameters
    ----------
    vertex : :class:`.Vertex`
        The vertex to test.

    position : :class:`numpy.ndarray`
        The position of the new vertex.

    Returns
    -------
    None : :class:`NoneType`

    """

    clone = vertex.with_position(position)
    assert np.all(np.equal(clone.get_position(), position))



import itertools as it

import numpy as np

from ..utilities import normalize_ids


def test_get_atomic_positions(case_data, get_atom_ids):
    """
    Test :meth:`.Molecule.get_atomic_positions`.

    Parameters
    ----------
    case_data : :class:`.CaseData`.
        A test case. Holds the molecule to test and the correct
        atomic positions.

    get_atom_ids : :class:`callable`
        Takes a single parameter, `molecule` and returns a valid
        `atom_ids` parameter for
        :meth:`.Molecule.get_atomic_positions`. This allows the testing
        of different values of this parameter.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_atomic_positions(
        molecule=case_data.molecule,
        position_matrix=case_data.position_matrix,
        get_atom_ids=get_atom_ids,
    )


def _test_get_atomic_positions(
    molecule,
    position_matrix,
    get_atom_ids,
):
    """
    Test :meth:`.Molecule.get_atomic_positions`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    position_matrix : :class:`numpy.ndarray`
        The correct atomic positions.

    get_atom_ids : :class:`callable`
        Takes a single parameter, `molecule` and returns a valid
        `atom_ids` parameter for
        :meth:`.Molecule.get_atomic_positions`. This allows the testing
        of different values of this parameter.

    Returns
    -------
    None : :class:`NoneType`

    """

    for atom_id, position in it.zip_longest(
        normalize_ids(molecule, get_atom_ids(molecule)),
        molecule.get_atomic_positions(get_atom_ids(molecule)),
    ):
        assert np.allclose(
            a=position,
            b=position_matrix[atom_id],
            atol=1e-6,
        )



import typing
from collections.abc import Iterable

import numpy as np

from stk._internal.ea.mutation.mutator import MoleculeMutator
from stk._internal.ea.mutation.record import MutationRecord

T = typing.TypeVar("T")


class RandomMutator:
    """
    Use some other mutator at random.

    Examples:
        *Use One of Several Mutators at Random*

        .. testcode:: use-one-of-several-mutators-at-random

            import stk

            random_seed = 12
            mutator = stk.RandomMutator(
                mutators=(
                    stk.RandomBuildingBlock(
                        building_blocks=(
                            stk.BuildingBlock(
                                smiles='BrCCBr',
                                functional_groups=[stk.BromoFactory()],
                            ),
                            stk.BuildingBlock(
                                smiles='BrCCCBr',
                                functional_groups=[stk.BromoFactory()],
                            ),
                        ),
                        is_replaceable=lambda building_block: True,
                        random_seed=random_seed,
                    ),
                    stk.SimilarBuildingBlock(
                        building_blocks=(
                            stk.BuildingBlock(
                                smiles='BrCCCCBr',
                                functional_groups=[stk.BromoFactory()],
                            ),
                            stk.BuildingBlock(
                                smiles='BrCCCCCBr',
                                functional_groups=[stk.BromoFactory()],
                            ),
                        ),
                        is_replaceable=lambda building_block: True,
                        random_seed=random_seed,
                    ),
                    stk.RandomBuildingBlock(
                        building_blocks=(
                            stk.BuildingBlock(
                                smiles='BrCCNCBr',
                                functional_groups=[stk.BromoFactory()],
                            ),
                            stk.BuildingBlock(
                                smiles='BrCNCBr',
                                functional_groups=[stk.BromoFactory()],
                            ),
                        ),
                        is_replaceable=lambda building_block: True,
                        random_seed=random_seed,
                    ),
                ),
            )
            building_block = stk.BuildingBlock(
                smiles='BrCNNCBr',
                functional_groups=[stk.BromoFactory()],
            )
            record = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            # Use one of the component mutators at random.
            mutation_record1 = mutator.mutate(record)
            # A different mutator may get selected at random the second,
            # third, etc, time.
            mutation_record2 = mutator.mutate(record)

        .. testcode:: use-one-of-several-mutators-at-random
            :hide:

            _smiles = stk.Smiles()
            assert _smiles.get_key(record.get_molecule()) == 'BrCNNCCNNCBr'

            _molecule1 = (
                mutation_record1.get_molecule_record().get_molecule()
            )
            assert _smiles.get_key(_molecule1) != 'BrCNNCNNCBr'

            _molecule2 = (
                mutation_record2.get_molecule_record().get_molecule()
            )
            assert _smiles.get_key(_molecule2) != 'BrCNNCNNCBr'
    """

    def __init__(
        self,
        mutators: Iterable[MoleculeMutator[T]],
        weights: Iterable[float] | None = None,
        random_seed: int | np.random.Generator | None = None,
    ) -> None:
        """
        Parameters:
            mutators (list[MoleculeMutator[T]]):
                A selection of mutators, each time :meth:`.mutate`
                is called, one will be selected at random to perform
                the mutation.

            weights (list[float] | None):
                For each mutator, the probability that it will be chosen
                whenever :meth:`.mutate` is called.
                If ``None`` all `mutators` will have equal chance of being
                selected.

            random_seed:
                The random seed to use.
        """
        if weights is not None:
            weights = tuple(weights)

        if random_seed is None or isinstance(random_seed, int):
            random_seed = np.random.default_rng(random_seed)

        self._mutators = tuple(mutators)
        self._weights = weights
        self._generator = random_seed

    def mutate(self, record: T) -> MutationRecord[T] | None:
        """
        Return a mutant of `record`.

        Parameters:
            record:
                The molecule to be mutated.

        Returns:
            A record of the mutation or ``None``
            if `record` cannot be mutated.
        """

        mutator = self._generator.choice(
            a=self._mutators,  # type: ignore
            p=self._weights,
        )
        return mutator.mutate(record)



from stk._internal.topology_graphs.polymer.linear import Linear
from stk._internal.topology_graphs.polymer.vertices import (
    HeadVertex,
    LinearVertex,
    TailVertex,
    TerminalVertex,
    UnaligningVertex,
)

__all__ = [
    "Linear",
    "TerminalVertex",
    "HeadVertex",
    "TailVertex",
    "LinearVertex",
    "UnaligningVertex",
]



import numpy as np

from .utilities import is_clone


def test_with_scale(vertex, scale_data):
    """
    Test :meth:`.Vertex.with_scale`.

    Parameters
    ----------
    vertex : :class:`.Vertex`
        The vertex to test.

    scale_data : :class:`.ScaleData`
        The parameter for :meth:`.Vertex.with_scale` to test.

    Returns
    -------
    None : :class:`NoneType`

    """

    # Save a clone to check immutability.
    clone = vertex.clone()
    _test_with_scale(
        vertex=vertex,
        start=scale_data.start,
        scale=scale_data.scale,
        target=scale_data.target,
    )
    is_clone(vertex, clone)


def _test_with_scale(vertex, start, scale, target):
    """
    Test :meth:`.Vertex.with_scale`.

    Parameters
    ----------
    vertex : :class:`.Vertex`
        The vertex to test.

    start : :class:`numpy.ndarray`
        The starting position of `vertex`.

    scale : :class:`float` or :class:`numpy.ndarray`
        The scale to apply to `vertex`.

    target : :class:`numpy.ndarray`
        The correct position of the new vertex.

    Returns
    -------
    None : :class:`NoneType`

    """

    clone = vertex.with_position(start).with_scale(scale)
    assert np.allclose(
        a=clone.get_position(),
        b=target,
        atol=1e-16,
    )



import pytest
from pytest_lazyfixture import lazy_fixture

from ...case_data import CaseData

# Fixtures need to be visible for lazy_fixture() calls.
from .internal_reaction import *  # noqa
from .ncore import *  # noqa


@pytest.fixture(
    params=(
        lazy_fixture("small_ncore"),
        lazy_fixture("small_internal"),
    ),
)
def small(request: pytest.FixtureRequest) -> CaseData:
    return request.param



from collections.abc import Iterable, Sequence
from typing import Any


class HashableDict(dict):
    def __hash__(self):
        return hash(
            (
                frozenset(self),
                frozenset(_to_hashable(list(self.values()))),
            )
        )

    def __eq__(self, other):
        return super().__eq__(other)


def _to_hashable(item):
    if isinstance(item, list):
        return tuple(_to_hashable(subitem) for subitem in item)
    else:
        return item


def get_any_value(
    mapping: dict[str, Sequence[dict[Any, Any]]],
    keys: Iterable[str],
) -> dict[Any, Any] | None:
    """
    Return any value in `mapping` for any of the `keys`.

    This function will only return a value if it is a non-empty
    sequence. The function will not return the sequence itself, but
    its first member.

    Parameters:

        mapping:
            A mapping from which a value is to be extracted.

        keys:
            The keys which are used to look up values in `mapping`.

    Returns:

        The value of the first key found in `mapping` wll be returned.
        If none of the `keys` are found in `mapping`, ``None`` will be
        returned.

    """

    for key in keys:
        if key in mapping and len(mapping[key]) > 0:
            return mapping[key][0]
    return None



import pytest
import stk

from .case_data import CaseData


@pytest.fixture(
    params=(
        lambda: CaseData(molecule=stk.BuildingBlock("NCCN"), num_bonds=11),
    ),
)
def case_data(request) -> CaseData:
    """
    A :class:`.CaseData` instance.

    """

    return request.param()



import numpy as np
import pytest
from pytest_lazyfixture import lazy_fixture

import stk

from ...case_data import CaseData


@pytest.fixture(
    params=(
        lazy_fixture("substituent"),
        lazy_fixture("core"),
        lazy_fixture("single"),
    ),
)
def case_data(request: pytest.FixtureRequest) -> CaseData:
    return request.param


@pytest.fixture
def core(
    id: int,
    position: tuple[float, float, float] | np.ndarray,
) -> CaseData:
    return CaseData(
        vertex=stk.small.CoreVertex(
            id=id,
            position=position,
        ),
        id=id,
        position=position,
        cell=np.array([0, 0, 0]),
    )


@pytest.fixture
def single(
    id: int,
    position: tuple[float, float, float] | np.ndarray,
) -> CaseData:
    return CaseData(
        vertex=stk.small.SingleVertex(
            id=id,
            position=position,
        ),
        id=id,
        position=position,
        cell=np.array([0, 0, 0]),
    )


@pytest.fixture
def substituent(
    id: int,
    position: tuple[float, float, float] | np.ndarray,
) -> CaseData:
    return CaseData(
        vertex=stk.small.SubstituentVertex(
            id=id,
            position=position,
        ),
        id=id,
        position=position,
        cell=np.array([0, 0, 0]),
    )


@pytest.fixture(params=(0,))
def id(request: pytest.FixtureRequest) -> int:
    return request.param



from typing import Callable

import pytest
import stk


def get_id_map_0(
    functional_group: stk.FunctionalGroup,
) -> dict[int, int]:
    """
    Get an id_map with a single atom.

    """

    # Make sure new_id is always valid, by making 1 larger than the
    # biggest on in the functional group. This prevents two atoms in
    # the functional group from having the same id.
    new_id = max(functional_group.get_atom_ids()) + 1
    ids = (new_id,)
    return dict(zip(functional_group.get_atom_ids(), ids))


def get_id_map_1(
    functional_group: stk.FunctionalGroup,
) -> dict[int, int]:
    new_id = max(functional_group.get_atom_ids()) + 1
    num_atoms = sum(1 for _ in functional_group.get_atoms())
    ids = range(new_id, new_id + num_atoms)
    return dict(zip(functional_group.get_atom_ids(), ids))


@pytest.fixture(
    params=[
        lambda functional_group: {},
        get_id_map_0,
        get_id_map_1,
    ],
)
def get_id_map(
    request,
) -> Callable[[stk.FunctionalGroup], dict[int, int]]:
    """
    Return a valid `id_map` parameter for a functional group.

    Parameters:

        functional_group:
            A functional group being tested, for which a valid
            `id_map` parameter needs to created.

    Returns:

        A valid `id_map` parameter for calling
        :meth:`~.FunctionalGroup.with_ids` on `functional_group`.

    """

    return request.param



import numpy as np
import pytest
import stk
from pytest_lazyfixture import lazy_fixture

# Fixtures must by visible for lazy_fixture() calls.
from .fixtures import *  # noqa


@pytest.fixture(
    params=(
        lazy_fixture("center"),
        lazy_fixture("head_1"),
        lazy_fixture("head_2"),
        lazy_fixture("head_3"),
        lazy_fixture("tail_1"),
        lazy_fixture("tail_2"),
        lazy_fixture("tail_3"),
    ),
)
def case_data(request):
    """
    A :class:`.CaseData` instance.

    """

    return request.param


@pytest.fixture(
    params=(
        lambda: stk.BuildingBlock(
            smiles="BrCCNCBr",
            functional_groups=[stk.BromoFactory()],
        ),
    ),
)
def building_block_2(request) -> stk.BuildingBlock:
    """
    A :class:`.BuildingBlock` with 2 functional groups.

    """

    return request.param()


@pytest.fixture(
    params=(
        lambda: stk.BuildingBlock(
            smiles="BrCC",
            functional_groups=[stk.BromoFactory()],
        ),
    ),
)
def building_block_1(request) -> stk.BuildingBlock:
    """
    A :class:`.BuildingBlock` with 1 functional group.

    """

    return request.param()


@pytest.fixture(params=(True, False))
def flip(request):
    return request.param


@pytest.fixture(
    params=(
        [0, 0, 0],
        [1, 2, -20],
    ),
)
def position(request):
    return np.array(request.param, dtype=np.float64)



from stk._internal.bond import Bond
from stk._internal.molecule import Molecule


class BondInfo:
    """
    Holds additional info about :class:`.ConstructedMolecule` bonds.
    """

    def __init__(
        self,
        bond: Bond,
        building_block: Molecule | None,
        building_block_id: int | None,
    ) -> None:
        """
        Parameters:
            bond:
                The bond about which information is held.

            building_block:
                The building block from which the bond originates.
                Can be ``None``, if the bond was not part of a building
                block, but was added by the construction process instead.

            building_block_id:
                A unique id for each :class:`.Molecule` placed during
                the construction of the :class:`.ConstructedMolecule`. As a
                single :class:`.Molecule` can be placed multiple times
                during construction, the `building_block_id` allows
                the user to distinguish between each placement. Can be
                ``None``, if the bond was not part of a building block, but
                was added by the construction process instead.
        """
        self._bond = bond
        self._building_block = building_block
        self._building_block_id = building_block_id

    def get_bond(self) -> Bond:
        """
        Get the bond about which information is held.

        Returns:
            The bond.
        """
        return self._bond

    def get_building_block(self) -> Molecule | None:
        """
        Get the building block from which the bond originates.

        Returns:
            The building block or ``None``, if the bond was not
            originally found in a building block, but was added by
            the construction process instead.
        """
        return self._building_block

    def get_building_block_id(self) -> int | None:
        """
        Get the id of the bond's building block.

        A unique id for each :class:`.Molecule` placed during
        the construction of the :class:`.ConstructedMolecule`. As a
        single :class:`.Molecule` can be placed multiple times
        during construction, the building block id  allows
        the user to distinguish between each placement.

        Returns:
            The id, if the bond was not originally found in a
            building block, but was added by the construction
            process instead.
        """
        return self._building_block_id



"""
COF Edge
========

"""

from stk._internal.topology_graphs.edge import Edge


class CofEdge(Edge):
    """
    A :class:`.Cof` edge.

    """

    def __init__(
        self,
        parent_id,
        id,
        vertex1,
        vertex2,
        periodicity=(0, 0, 0),
        position=None,
    ):
        """
        Initialize a :class:`.CofEdge`.

        Parameters
        ----------
        parent_id : :class:`int`
            The id of the edge in the (1, 1, 1) unit cell, with which
            this one is periodic.

        id : :class:`int`
            The id of the edge.

        vertex1 : :class:`.Vertex`
            The first vertex the edge is connected to.

        vertex2 : :class:`.Vertex`
            The second vertex the edge is connected to.

        periodicity : :class:`tuple` of :class:`int`, optional
            The periodicity of the edge, when going from `vertex1` to
            `vertex2`. For example, if ``(0, 0, 0)`` the edge is not
            periodic, if ``(1, 0, -1)`` the edge is periodic going in
            the positive direction along the x axis, is not periodic
            across the y axis and is periodic in the negative direction
            along the z axis.

        position : :class:`numpy.ndarray`, optional
            The position of the edge, if ``None``, the midpoint of the
            vertices is used.

        """

        super().__init__(id, vertex1, vertex2, periodicity, position)
        self._parent_id = parent_id

    def get_parent_id(self):
        """
        Get the id of the *parent* edge.

        The *parent* edge is the edge in the (1, 1, 1) unit cell, with
        which this one is periodic.

        Returns
        -------
        :class:`int`
            The id of the parent edge.

        """

        return self._parent_id

    def clone(self):
        obj = super().clone()
        obj._parent_id = self._parent_id
        return obj



class MoleculeDatabase:
    """
    An abstract base class for storing and retrieving molecules.

    See Also
    --------
    :class:`.ConstructedMoleculeDatabase`
        If you need to store and retrieve
        :class:`.ConstructedMolecule` instances. You can put \
        :class:`.ConstructedMolecule` instances into a \
        :class:`.MoleculeDatabase`, however, you will only be able to \
        retrieve them as plain :class:`.Molecule` instances.

    Examples
    --------
    *Subclass Implementation*

    The source code of the subclasses, listed in
    :mod:`molecule_database <~.databases.molecule>`, can serve as
    good examples.

    """

    def put(self, molecule):
        """
        Put `molecule` into the database.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            The molecule to place into the database.

        Returns
        -------
        None : :class:`NoneType`

        """

        raise NotImplementedError()

    def get(self, key):
        """
        Get the molecule with `key` from the database.

        Parameters
        ----------
        key : :class:`object`
            The key of a molecule, which is to be returned from the
            database.

        Returns
        -------
        :class:`.Molecule`
            The molecule held in the database under `key`.

        Raises
        ------
        :class:`KeyError`
            If `key` is not found in the database.

        """

        raise NotImplementedError()

    def get_all(self):
        """
        Get all molecules in the database.

        Yields
        ------
        :class:`.Molecule`
            A molecule in the database.

        """

        raise NotImplementedError()



from stk import BondInfo


def check_bond_infos(old_state, new_state, building_blocks):
    old_infos = tuple(old_state.get_bond_infos())
    new_infos = tuple(new_state.get_bond_infos())
    num_added = len(new_infos) - len(old_infos)
    assert num_added == sum(bb.get_num_bonds() for bb in building_blocks)

    for info1, info2 in zip(
        # Take just the newly added infos.
        new_infos[len(old_infos) :],
        get_infos(old_state, new_state, building_blocks),
    ):
        is_equivalent_bond(info1.bond, info2.bond)
        assert info1.building_block is info2.building_block
        assert info1.building_block_index is info2.building_block_index


def get_infos(old_state, new_state, building_blocks):
    start_building_block_index = max(
        (info.building_block_index + 1 for info in old_state.get_bond_infos()),
        default=0,
    )
    start_atom_index = sum(1 for _ in old_state.get_atoms())
    for building_block_index, building_block in enumerate(
        building_blocks,
        start_building_block_index,
    ):
        atom_map = {
            atom.get_id(): atom.with_id(id_)
            for id_, atom in enumerate(
                building_block.get_atoms(),
                start_atom_index,
            )
        }
        start_atom_index += building_block.get_num_atoms()

        for bond in building_block.get_bonds():
            yield BondInfo(
                bond=bond.with_atoms(atom_map),
                building_block=building_block,
                building_block_index=building_block_index,
            )


def is_equivalent_bond(bond1, bond2):
    is_equivalent_atom(bond1.get_atom1(), bond2.get_atom1())
    is_equivalent_atom(bond1.get_atom2(), bond2.get_atom2())


def is_equivalent_atom(atom1, atom2):
    assert atom1.__class__ is atom2.__class__
    assert atom1.get_id() == atom2.get_id()
    assert atom1.get_charge() == atom2.get_charge()



class CaseData:
    """
    A test case.

    Attributes
    ----------
    key_maker : :class:`.MoleculeKeyMaker`
        The key maker to test.

    molecule : :class:`.Molecule`
        The molecule to test.

    key_name : :class:`str`
        The name of the key made by :attr:`.key_maker`.

    key : :class:`object`
        The correct key.

    """

    def __init__(
        self,
        key_maker,
        molecule,
        key_name,
        key,
    ):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        key_maker : :class:`.MoleculeKeyMaker`
            The key maker to test.

        molecule : :class:`.Molecule`
            The molecule to test.

        key_name : :class:`str`
            The name of the key made by `key_maker`.

        key : :class:`object`
            The correct key.

        """

        self.key_maker = key_maker
        self.molecule = molecule
        self.key_name = key_name
        self.key = key



from .cage import *  # noqa
from .cof import *  # noqa
from .host_guest import *  # noqa
from .macrocycle import *  # noqa
from .metal_complex import *  # noqa
from .polymer import *  # noqa
from .rotaxane import *  # noqa
from .small import *  # noqa



import pytest
import stk

from ..case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda: CaseData(
            building_block=stk.BuildingBlock.init_from_molecule(
                molecule=stk.BuildingBlock("Br[C+2][C+2]Br"),
            ),
            functional_groups=(),
            known_repr="BuildingBlock('Br[C+2][C+2]Br', ())",
            core_atom_ids=(0, 1, 2, 3),
            placer_ids=(0, 1, 2, 3),
        ),
        lambda: CaseData(
            building_block=stk.BuildingBlock.init_from_molecule(
                molecule=stk.BuildingBlock("Br[C+2][C+2]Br"),
                functional_groups=[stk.BromoFactory()],
            ),
            functional_groups=(
                stk.Bromo(
                    bromine=stk.Br(0),
                    atom=stk.C(1, 2),
                    bonders=(stk.C(1, 2),),
                    deleters=(stk.Br(0),),
                ),
                stk.Bromo(
                    bromine=stk.Br(3),
                    atom=stk.C(2, 2),
                    bonders=(stk.C(2, 2),),
                    deleters=(stk.Br(3),),
                ),
            ),
            known_repr=(
                "BuildingBlock('Br[C+2][C+2]Br', (Bromo(Br(0), C(1, charge=2),"
                " bonders=(C(1, charge=2),), deleters=(Br(0),)), Bromo(Br(3), "
                "C(2, charge=2), bonders=(C(2, charge=2),), deleters=(Br(3),))"
                "))"
            ),
            core_atom_ids=(1, 2),
            placer_ids=(1, 2),
        ),
        lambda: CaseData(
            building_block=stk.BuildingBlock.init_from_molecule(
                molecule=stk.BuildingBlock("Br[C+2][C+2]Br"),
                placer_ids=(1, 2),
            ),
            functional_groups=(),
            known_repr="BuildingBlock('Br[C+2][C+2]Br', ())",
            core_atom_ids=(0, 1, 2, 3),
            placer_ids=(1, 2),
        ),
        lambda: CaseData(
            building_block=stk.BuildingBlock.init_from_molecule(
                molecule=stk.BuildingBlock("Br[C+2][C+2]Br"),
                functional_groups=[stk.BromoFactory()],
                placer_ids=(0, 3),
            ),
            functional_groups=(
                stk.Bromo(
                    bromine=stk.Br(0),
                    atom=stk.C(1, 2),
                    bonders=(stk.C(1, 2),),
                    deleters=(stk.Br(0),),
                ),
                stk.Bromo(
                    bromine=stk.Br(3),
                    atom=stk.C(2, 2),
                    bonders=(stk.C(2, 2),),
                    deleters=(stk.Br(3),),
                ),
            ),
            known_repr=(
                "BuildingBlock('Br[C+2][C+2]Br', (Bromo(Br(0), C(1, charge=2),"
                " bonders=(C(1, charge=2),), deleters=(Br(0),)), Bromo(Br(3), "
                "C(2, charge=2), bonders=(C(2, charge=2),), deleters=(Br(3),))"
                "))"
            ),
            core_atom_ids=(1, 2),
            placer_ids=(0, 3),
        ),
        lambda: CaseData(
            building_block=stk.BuildingBlock.init_from_molecule(
                molecule=stk.BuildingBlock("Br[C+2][C+2]Br"),
                functional_groups=[stk.IodoFactory()],
            ),
            functional_groups=(),
            known_repr="BuildingBlock('Br[C+2][C+2]Br', ())",
            core_atom_ids=(0, 1, 2, 3),
            placer_ids=(0, 1, 2, 3),
        ),
    ),
)
def init_from_molecule(request) -> CaseData:
    return request.param()



import pytest
from pytest_lazyfixture import lazy_fixture

# Fixtures need to be visible for lazy_fixture() calls.
from .fixtures import *  # noqa


@pytest.fixture(
    params=(
        lazy_fixture("host"),
        lazy_fixture("guest"),
    ),
)
def case_data(request):
    """
    A :class:`.CaseData` instance.

    """

    return request.param



import pytest
import stk

from ....case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cage.FivePlusTen(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="BrC1=C(Br)[C+]=N1",
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles=(
                                "Br[C+]1C2(Br)[C+]=N[C+]2[C+](Br)[C+]("
                                "Br)[C+2]1"
                            ),
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                ),
            ),
            smiles=(
                "[C+]1=NC2=C1[C+]1[C+2][C+]3C4=C([C+]=N4)C45[C+]=N[C+]"
                "4[C+]4C6=C(N=[C+]6)[C+]3[C+]3N=[C+]C13C1=C(N=[C+]1)[C"
                "+]1[C+]2[C+2][C+]2C3=C(N=[C+]3)[C+]3[C+2][C+]6C7=C(N="
                "[C+]7)[C+]7[C+2][C+]8C9=C(N=[C+]9)[C+]5[C+2][C+]4C4=C"
                "([C+]=N4)C84[C+]=N[C+]4[C+]7C4=C([C+]=N4)C64[C+]=N[C+"
                "]4[C+]3C3=C([C+]=N3)C23[C+]=N[C+]13"
            ),
            name=name,
        ),
    ),
)
def cage_five_plus_ten(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



def test_get_building_block(case_data):
    """
    Test :meth:`.AtomInfo.get_building_block`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the atom info to test and the building block
        it should be holding.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert case_data.atom_info.get_building_block() is case_data.building_block



import numpy as np
import pytest
import stk

from ..case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda: CaseData(
            building_block=stk.BuildingBlock.init(
                atoms=(stk.Br(0), stk.C(1, 2), stk.C(2, 2), stk.Br(3)),
                bonds=(
                    stk.Bond(stk.Br(0), stk.C(1, 2), 1),
                    stk.Bond(stk.C(1, 2), stk.C(2, 2), 1),
                    stk.Bond(stk.C(2, 2), stk.Br(3), 1),
                ),
                position_matrix=np.array(
                    [
                        [0.0, 0.0, 0.0],
                        [1.0, 1.0, 1.0],
                        [2.0, 2.0, 2.0],
                        [3.0, 3.0, 3.0],
                    ]
                ),
            ),
            functional_groups=(),
            known_repr="BuildingBlock('Br[C+2][C+2]Br', ())",
            core_atom_ids=(0, 1, 2, 3),
            placer_ids=(0, 1, 2, 3),
        ),
        lambda: CaseData(
            building_block=stk.BuildingBlock.init(
                atoms=(stk.Br(0), stk.C(1, 2), stk.C(2, 2), stk.Br(3)),
                bonds=(
                    stk.Bond(stk.Br(0), stk.C(1, 2), 1),
                    stk.Bond(stk.C(1, 2), stk.C(2, 2), 1),
                    stk.Bond(stk.C(2, 2), stk.Br(3), 1),
                ),
                position_matrix=np.array(
                    [
                        [0.0, 0.0, 0.0],
                        [1.0, 1.0, 1.0],
                        [2.0, 2.0, 2.0],
                        [3.0, 3.0, 3.0],
                    ]
                ),
                functional_groups=[stk.BromoFactory()],
            ),
            functional_groups=(
                stk.Bromo(
                    bromine=stk.Br(0),
                    atom=stk.C(1, 2),
                    bonders=(stk.C(1, 2),),
                    deleters=(stk.Br(0),),
                ),
                stk.Bromo(
                    bromine=stk.Br(3),
                    atom=stk.C(2, 2),
                    bonders=(stk.C(2, 2),),
                    deleters=(stk.Br(3),),
                ),
            ),
            known_repr=(
                "BuildingBlock('Br[C+2][C+2]Br', (Bromo(Br(0), C(1, charge=2),"
                " bonders=(C(1, charge=2),), deleters=(Br(0),)), Bromo(Br(3), "
                "C(2, charge=2), bonders=(C(2, charge=2),), deleters=(Br(3),))"
                "))"
            ),
            core_atom_ids=(1, 2),
            placer_ids=(1, 2),
        ),
        lambda: CaseData(
            building_block=stk.BuildingBlock.init(
                atoms=(stk.Br(0), stk.C(1, 2), stk.C(2, 2), stk.Br(3)),
                bonds=(
                    stk.Bond(stk.Br(0), stk.C(1, 2), 1),
                    stk.Bond(stk.C(1, 2), stk.C(2, 2), 1),
                    stk.Bond(stk.C(2, 2), stk.Br(3), 1),
                ),
                position_matrix=np.array(
                    [
                        [0.0, 0.0, 0.0],
                        [1.0, 1.0, 1.0],
                        [2.0, 2.0, 2.0],
                        [3.0, 3.0, 3.0],
                    ]
                ),
                placer_ids=(1, 2),
            ),
            functional_groups=(),
            known_repr="BuildingBlock('Br[C+2][C+2]Br', ())",
            core_atom_ids=(0, 1, 2, 3),
            placer_ids=(1, 2),
        ),
        lambda: CaseData(
            building_block=stk.BuildingBlock.init(
                atoms=(stk.Br(0), stk.C(1, 2), stk.C(2, 2), stk.Br(3)),
                bonds=(
                    stk.Bond(stk.Br(0), stk.C(1, 2), 1),
                    stk.Bond(stk.C(1, 2), stk.C(2, 2), 1),
                    stk.Bond(stk.C(2, 2), stk.Br(3), 1),
                ),
                position_matrix=np.array(
                    [
                        [0.0, 0.0, 0.0],
                        [1.0, 1.0, 1.0],
                        [2.0, 2.0, 2.0],
                        [3.0, 3.0, 3.0],
                    ]
                ),
                functional_groups=[stk.BromoFactory()],
                placer_ids=(0, 3),
            ),
            functional_groups=(
                stk.Bromo(
                    bromine=stk.Br(0),
                    atom=stk.C(1, 2),
                    bonders=(stk.C(1, 2),),
                    deleters=(stk.Br(0),),
                ),
                stk.Bromo(
                    bromine=stk.Br(3),
                    atom=stk.C(2, 2),
                    bonders=(stk.C(2, 2),),
                    deleters=(stk.Br(3),),
                ),
            ),
            known_repr=(
                "BuildingBlock('Br[C+2][C+2]Br', (Bromo(Br(0), C(1, charge=2),"
                " bonders=(C(1, charge=2),), deleters=(Br(0),)), Bromo(Br(3), "
                "C(2, charge=2), bonders=(C(2, charge=2),), deleters=(Br(3),))"
                "))"
            ),
            core_atom_ids=(1, 2),
            placer_ids=(0, 3),
        ),
        lambda: CaseData(
            building_block=stk.BuildingBlock.init(
                atoms=(stk.Br(0), stk.C(1, 2), stk.C(2, 2), stk.Br(3)),
                bonds=(
                    stk.Bond(stk.Br(0), stk.C(1, 2), 1),
                    stk.Bond(stk.C(1, 2), stk.C(2, 2), 1),
                    stk.Bond(stk.C(2, 2), stk.Br(3), 1),
                ),
                position_matrix=np.array(
                    [
                        [0.0, 0.0, 0.0],
                        [1.0, 1.0, 1.0],
                        [2.0, 2.0, 2.0],
                        [3.0, 3.0, 3.0],
                    ]
                ),
                functional_groups=[stk.IodoFactory()],
            ),
            functional_groups=(),
            known_repr="BuildingBlock('Br[C+2][C+2]Br', ())",
            core_atom_ids=(0, 1, 2, 3),
            placer_ids=(0, 1, 2, 3),
        ),
    ),
)
def init(request) -> CaseData:
    return request.param()



import pytest
import stk


@pytest.fixture
def construction_state(building_block_vertices):
    return stk.ConstructionState(
        building_block_vertices=building_block_vertices,
        edges=get_edges(building_block_vertices),
        scale=1,
    )


def get_edges(building_block_vertices):
    vertices = (
        vertex
        for vertices in building_block_vertices.values()
        for vertex in vertices
    )
    vertex1 = next(vertices)
    for id_, vertex2 in enumerate(vertices):
        yield stk.Edge(id_, vertex1, vertex2)
        vertex1 = vertex2


@pytest.fixture
def building_block_vertices(building_block1, building_block2):
    return {
        building_block1: (
            stk.Vertex(0, [0, 0, 0]),
            stk.Vertex(1, [10, 0, 0]),
        ),
        building_block2: (
            stk.Vertex(2, [20, 0, 0]),
            stk.Vertex(3, [30, 0, 0]),
        ),
    }


@pytest.fixture(
    params=(stk.BuildingBlock("BrCC", [stk.BromoFactory()]),),
)
def building_block1(request):
    return request.param


@pytest.fixture(
    params=(stk.BuildingBlock("BrCCBr", [stk.BromoFactory()]),),
)
def building_block2(request):
    return request.param



import itertools as it

from tests.utilities import is_equivalent_bond


def test_get_bonds(case_data):
    """
    Test :meth:`.Molecule.get_bonds`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        The test case. Holds the molecule to test and the correct
        bonds.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_bonds(case_data.molecule, case_data.bonds)


def _test_get_bonds(molecule, bonds):
    """
    Test :meth:`.Molecule.get_bonds`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    bonds : :class:`tuple` of :class:`.Bond`
        The correct bonds of `molecule`.

    Returns
    -------
    None : :class:`NoneType`

    """

    for bond1, bond2 in it.zip_longest(molecule.get_bonds(), bonds):
        is_equivalent_bond(bond1, bond2)



import itertools as it

from .case_data import CaseData


def test_get_core_atom_ids(case_data: CaseData) -> None:
    """
    Test :meth:`.BuildingBlock.get_core_atom_ids`.

    Parameters:
        case_data:
            A test case. Holds the building block to test and the correct
            core atom ids.

    """

    for id1, id2 in it.zip_longest(
        case_data.building_block.get_core_atom_ids(),
        case_data.core_atom_ids,
    ):
        assert id1 == id2



"""
Construction State
==================

"""

from .graph_state import GraphState
from .molecule_state import MoleculeState


class ConstructionState:
    """
    The state of the molecule and topology graph under construction.

    """

    def __init__(
        self,
        building_block_vertices,
        edges,
        lattice_constants=(),
    ):
        """
        Initialize a :class:`.ConstructionState` instance.

        Parameters
        ----------
        building_block_vertices : :class:`dict`
            Maps each :class:`.BuildingBlock` to be placed, to a
            :class:`tuple` of :class:`.Vertex` instances, on which
            it should be placed.

        edges : :class:`tuple` of :class:`.Edge`
            The edges of the topology graph.

        lattice_constants : :class:`tuple`, optional
            A :class:`numpy.ndarray` for each lattice constant.
            Can be an empty :class:`tuple` if the topology graph is
            not periodic.

        """

        self._graph_state = GraphState(
            building_block_vertices=building_block_vertices,
            edges=edges,
            lattice_constants=lattice_constants,
        )
        self._molecule_state = MoleculeState()

    def clone(self):
        """
        Return a clone.

        Returns
        -------
        :class:`.ConstructionState`
            The clone. Has the same type as the original instance.

        """

        clone = self.__class__.__new__(self.__class__)
        clone._graph_state = self._graph_state
        clone._molecule_state = self._molecule_state
        return clone

    def _with_placement_results(
        self,
        vertices,
        edges,
        building_blocks,
        results,
    ):
        """
        Modify the instance.

        """

        self._molecule_state = self._molecule_state.with_placement_results(
            vertices=vertices,
            edges=edges,
            building_blocks=building_blocks,
            results=results,
        )
        return self

    def with_placement_results(
        self,
        vertices,
        edges,
        building_blocks,
        results,
    ):
        """
        Return a clone holding the placement results.

        Parameters
        ----------
        vertices : :class:`tuple` of :class:`.Vertex`
            The vertices used for placement.

        edges : :class:`tuple`
            For each vertex in `vertices`, a :class:`tuple` of
            :class:`.Edge` instances connected to it.

        building_blocks : :class:`tuple` of :class:`.BuildingBlock`
            For each vertex in `vertices`, the building block placed
            on it.

        results : :class:`tuple` of :class:`._PlacementResult`
            For every vertex in `vertices`, the result of the
            placement.

        Returns
        -------
        :class:`.ConstructionState`
            The clone holding the placement results. Has the same
            type as the original instance.

        """

        return self.clone()._with_placement_results(
            vertices=vertices,
            edges=edges,
            building_blocks=building_blocks,
            results=results,
        )

    def get_lattice_constants(self):
        """
        Get the lattice constants of the state.

        Returns
        -------
        :class:`tuple` of :class:`numpy.ndarray`
            The lattice constants.

        """

        return self._graph_state.get_lattice_constants()

    def get_building_block(self, vertex_id):
        """
        Get the building block to be placed on a given vertex.

        Parameters
        ----------
        vertex_id : :class:`int`
            The id of the vertex, on which the building block is to
            be placed.

        Returns
        -------
        :class:`.BuildingBlock`
            The building block.

        """

        return self._graph_state.get_building_block(vertex_id)

    def get_vertices(self, vertex_ids):
        """
        Get the building block to be placed on a given vertex.

        Parameters
        ----------
        vertex_id : :class:`int`
            The id of the vertex, on which the building block is to
            be placed.

        Returns
        -------
        :class:`.BuildingBlock`
            The building block.

        """

        yield from self._graph_state.get_vertices(vertex_ids)

    def get_num_vertices(self):
        """
        Get the number of vertices in the topology graph.

        Returns
        -------
        :class:`int`
            The number of vertices in the topology graph.

        """

        return self._graph_state.get_num_vertices()

    def get_edge(self, edge_id):
        """
        Get an edge.

        Parameters
        ----------
        edge_id : :class:`int`
            The id of an edge.

        Returns
        -------
        :class:`.Edge`
            An edge.

        """

        return self._graph_state.get_edge(edge_id)

    def get_num_edges(self):
        """
        Get the number of edges in the topology graph.

        Returns
        -------
        :class:`int`
            The number of edges.

        """

        return self._graph_state.get_num_edges()

    def get_edges(self, vertex_id):
        """
        Get the edges connected to a vertex.

        Parameters
        ----------
        vertex_id : :class:`int`
            The id of a vertex.

        Returns
        -------
        :class:`tuple` of :class:`.Edge`
            The connected edges.

        """

        return self._graph_state.get_edges(vertex_id)

    def get_edge_group_functional_groups(self, edge_group):
        """
        Yield the functional groups associated with `edge_group`.

        Parameters
        ----------
        edge_group : :class:`.EdgeGroup`
            The edge group, whose functional groups are desired.

        Yields
        ------
        :class:`.FunctionalGroup`
            A functional group which belongs to `edge_group`.

        """

        yield from (
            self._molecule_state.get_edge_group_functional_groups(
                edge_group=edge_group,
            )
        )

    def _with_reaction_results(self, reactions, results):
        """
        Modify the instance.

        """

        self._molecule_state = self._molecule_state.with_reaction_results(
            reactions=reactions,
            results=results,
        )
        return self

    def with_reaction_results(self, reactions, results):
        """
        Return a clone holding the reaction results.

        Parameters
        ----------
        reactions : :class:`tuple` of :class:`.Reaction`
            The reactions.

        results : :class:`.ReactionResult`
            For each reaction in `reactions`, its result.

        Returns
        -------
        :class:`.ConstructionState`
            The clone holding the reaction results. Has the same type
            as the original instance.

        """

        return self.clone()._with_reaction_results(reactions, results)

    def _with_lattice_constants(self, lattice_constants):
        """
        Modify the instance.

        """

        self._graph_state = self._graph_state.with_lattice_constants(
            lattice_constants=lattice_constants,
        )
        return self

    def with_lattice_constants(self, lattice_constants):
        """
        Return a clone holding the `lattice_constants`.

        Parameters
        ----------
        lattice_constants : :class:`tuple` of :class:`numpy.ndarray`
            The lattice constants of the clone. Requires 3 arrays of
            size``(3, )``.

        Returns
        -------
        :class:`.ConstructionState`
            The clone holding the new lattice constants. Has the same
            type as the original instance.

        """

        return self.clone()._with_lattice_constants(lattice_constants)

    def _with_position_matrix(self, position_matrix):
        """
        Modify the instance.

        """

        self._molecule_state = self._molecule_state.with_position_matrix(
            position_matrix=position_matrix,
        )
        return self

    def with_position_matrix(self, position_matrix):
        """
        Return a clone holding the `position_matrix`.

        Parameters
        ----------
        position_matrix : :class:`numpy.ndarray`
            The position matrix of the clone. The shape of the matrix
            is ``(n, 3)``.

        Returns
        -------
        :class:`.ConstructionState`
            The clone holding the new position matrix. Has the same
            type as the original instance.

        """

        return self.clone()._with_position_matrix(position_matrix)

    def _with_vertices(self, vertices):
        """
        Modify the instance.

        """

        self._graph_state = self._graph_state.with_vertices(vertices)
        return self

    def with_vertices(self, vertices):
        """
        Returns a clone holding `vertices`.

        Parameters
        ----------
        vertices : :class:`iterable` of :class:`.Vertex`
            The vertices the clone should hold.

        Returns
        -------
        :class:`.ConstructionState`
            The clone. Has the same type as the original instance.

        """

        return self.clone()._with_vertices(vertices)

    def get_position_matrix(self):
        """
        Get the position matrix of the molecule being constructed.

        Returns
        -------
        :class:`numpy.ndarray`
            The position matrix.

        """

        return self._molecule_state.get_position_matrix()

    def get_atoms(self):
        """
        Yield the atoms of the molecule being constructed.

        Yields
        ------
        :class:`.Atom`
            An atom of the molecule being constructed.

        """

        yield from self._molecule_state.get_atoms()

    def get_bonds(self):
        """
        Yield the bonds of the molecule being constructed.

        Yields
        ------
        :class:`.Bond`
            A bond of the molecule being constructed.

        """

        yield from self._molecule_state.get_bonds()

    def get_atom_infos(self):
        """
        Yield the atom infos of the molecule being constructed.

        Yields
        ------
        :class:`.AtomInfo`
            An atom info of the molecule being constructed.

        """

        yield from self._molecule_state.get_atom_infos()

    def get_bond_infos(self):
        """
        Yield the bond infos of the molecule being constructed.

        Yields
        ------
        :class:`.BondInfo`
            The bond info of the molecule being constructed.

        """

        yield from self._molecule_state.get_bond_infos()

    def get_num_building_block(self, building_block):
        """
        Get the number of times `building_block` is present.

        Parameters
        ----------
        building_block : :class:`.BuildingBlock`
            The building block whose frequency in the topology graph
            is desired.

        Returns
        -------
        :class:`int`
            The number of times `building_block` is present in the
            topology graph.

        """

        return self._graph_state.get_num_building_block(building_block)

    def get_building_blocks(self):
        """
        Yield the building blocks.

        Building blocks are yielded in an order based on their
        position in the topology graph. For two equivalent
        topology graphs, but with different building blocks,
        equivalently positioned building blocks will be yielded at the
        same time.

        Yields
        ------
        :class:`.BuildingBlock`
            A building block of the topology graph.

        """

        yield from self._graph_state.get_building_blocks()



import itertools as it
import typing
from collections import Counter, abc
from functools import partial
from operator import getitem

import numpy as np

from stk._internal.building_block import BuildingBlock
from stk._internal.construction_result.construction_result import (
    ConstructionResult,
)
from stk._internal.construction_result.periodic import (
    PeriodicConstructionResult,
)
from stk._internal.optimizers.null import (
    NullOptimizer,
)
from stk._internal.optimizers.optimizer import Optimizer
from stk._internal.reaction_factories.generic_reaction_factory import (
    GenericReactionFactory,
)
from stk._internal.reaction_factories.reaction_factory import ReactionFactory
from stk._internal.topology_graphs.edge import (
    Edge,
)
from stk._internal.topology_graphs.edge_group import (
    EdgeGroup,
)
from stk._internal.topology_graphs.topology_graph.topology_graph import (
    TopologyGraph,
)
from stk._internal.topology_graphs.vertex import Vertex

from .edge import CofEdge
from .vertices import UnaligningVertex, _CofVertex


class UnoccupiedVertexError(Exception):
    """
    When a COF vertex is not occupied by a building block.

    """

    pass


class OverlyOccupiedVertexError(Exception):
    """
    When a COF vertex is occupied by more than one building block.

    """

    pass


class Cof(TopologyGraph):
    """
    Represents a COF topology graph.

    Notes:

        COF topologies are added by creating a subclass, which defines
        the :attr:`_vertex_prototypes` and :attr:`_edge_prototypes` class
        attributes.

    .. _cof-topology-graph-examples:

    Examples:

        *Subclass Implementation*

        The source code of the subclasses, listed in :mod:`~.cof.cof`,
        can serve as good examples.

        *Basic Construction*

        :class:`.Cof` instances can be made by providing the building
        block molecules and lattice size only (using :class:`.Honeycomb`
        as an example)

        .. testcode:: basic-construction

            import stk

            bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])
            bb2 = stk.BuildingBlock('BrCC(CBr)CBr', [stk.BromoFactory()])
            cof = stk.ConstructedMolecule(
                topology_graph=stk.cof.Honeycomb((bb1, bb2), (3, 3, 1)),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])
            bb2 = stk.BuildingBlock('BrCC(CBr)CBr', [stk.BromoFactory()])
            cof = stk.ConstructedMolecule(
                topology_graph=stk.cof.Honeycomb((bb1, bb2), (3, 3, 1)),
            )
            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        cof.get_atoms(),
                        cof.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=(
                            1
                            if bond.get_order() == 9
                            else bond.get_order()
                        ),
                    ) for bond in cof.get_bonds()
                ),
            )

        *Suggested Optimization*

        For :class:`.Cof` topologies, it is recommend to use the
        :class:`.Collapser` optimizer if using the nonperiodic form.
        For periodic systems, the :class:`.PeriodicCollapser` is
        recommended.

        .. testcode:: suggested-optimization

            import stk

            bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])
            bb2 = stk.BuildingBlock('BrCC(CBr)CBr', [stk.BromoFactory()])

            # Nonperiodic.
            cof = stk.ConstructedMolecule(
                topology_graph=stk.cof.Honeycomb(
                    building_blocks=(bb1, bb2),
                    lattice_size=(3, 3, 1),
                    # Setting scale_steps to False tends to lead to a
                    # better structure.
                    optimizer=stk.Collapser(scale_steps=False),
                ),
            )

            # Periodic.
            cof = stk.ConstructedMolecule(
                topology_graph=stk.cof.PeriodicHoneycomb(
                    building_blocks=(bb1, bb2),
                    lattice_size=(3, 3, 1),
                    optimizer=stk.PeriodicCollapser(),
                ),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])
            bb2 = stk.BuildingBlock('BrCC(CBr)CBr', [stk.BromoFactory()])

            cof = stk.ConstructedMolecule(
                topology_graph=stk.cof.Honeycomb(
                    building_blocks=(bb1, bb2),
                    lattice_size=(3, 3, 1),
                    # Setting scale_steps to False tends to lead to a
                    # better structure.
                    optimizer=stk.Collapser(scale_steps=False),
                ),
            )
            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        cof.get_atoms(),
                        cof.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=(
                            1
                            if bond.get_order() == 9
                            else bond.get_order()
                        ),
                    ) for bond in cof.get_bonds()
                ),
            )

        *Accessing the Periodic Information*

        When building periodic :class:`.Cof` instances, the periodic
        information, such as the unit cell, can be accessed if you use the
        :class:`.PeriodicConstructionResult` returned by calling
        :meth:`.Cof.construct`

        .. testcode:: accessing-the-periodic-information

            import stk

            bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])
            bb2 = stk.BuildingBlock('BrCC(CBr)CBr', [stk.BromoFactory()])

            topology_graph = stk.cof.PeriodicHoneycomb(
                building_blocks=(bb1, bb2),
                lattice_size=(3, 3, 1),
            )
            construction_result = topology_graph.construct()
            cof = stk.ConstructedMolecule.init_from_construction_result(
                construction_result=construction_result,
            )
            periodic_info = construction_result.get_periodic_info()
            cell_matrix = periodic_info.get_cell_matrix()
            # Can access all unit-cell parameters.
            a = periodic_info.get_a()
            b = periodic_info.get_b()
            c = periodic_info.get_c()
            alpha = periodic_info.get_alpha()
            beta = periodic_info.get_beta()
            gamma = periodic_info.get_gamma()
            # Write to .pdb file.
            writer = stk.PdbWriter()
            writer.write(
                molecule=cof,
                path='cof.pdb',
                periodic_info=periodic_info,
            )

        .. testcleanup:: accessing-the-periodic-information

            import os
            os.remove('cof.pdb')


        *Structural Isomer Construction*

        Different structural isomers of COFs can be made by using the
        `vertex_alignments` optional parameter

        .. testcode:: structural-isomer-construction

            import stk

            bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])
            bb2 = stk.BuildingBlock('BrCC(CBr)CBr', [stk.BromoFactory()])

            cof2 = stk.ConstructedMolecule(
                topology_graph=stk.cof.Honeycomb(
                    building_blocks=(bb1, bb2),
                    lattice_size=(3, 3, 1),
                    vertex_alignments={0: 2, 1: 1, 2: 1},
                ),
            )

        The parameter maps the id of a vertex to a number
        between 0 (inclusive) and the number of edges the vertex is
        connected to (exclusive). So a vertex connected to three edges
        can be mapped to ``0``, ``1`` or ``2``.

        By changing which edge each vertex is aligned with, a different
        structural isomer of the COF can be formed.

        *Multi-Building Block COF Construction*

        You can also build COFs with multiple building blocks, but,
        if you have multiple building blocks with the same number of
        functional groups, you have to assign each building block to the
        vertex you want to place it on

        .. testcode:: multi-building-block-cof-construction

            import stk

            bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])
            bb2 = stk.BuildingBlock('BrCNCBr', [stk.BromoFactory()])
            bb3 = stk.BuildingBlock('BrCC(CBr)CBr', [stk.BromoFactory()])
            bb4 = stk.BuildingBlock('BrCC(NCBr)CBr', [stk.BromoFactory()])

            cof = stk.ConstructedMolecule(
                topology_graph=stk.cof.Honeycomb(
                    # building_blocks is now a dict, which maps building
                    # blocks to the id of the vertices it should be placed
                    # on. You can use ranges to specify the ids.
                    building_blocks={
                        bb1: (2, 4, 7, 8, 9, 12, 13, 14, 17, 18, 19),
                        bb2: 3,
                        bb3: (0, 10, 11, 15, 16),
                        bb4: (1, 5, 6),
                    },
                    lattice_size=(2, 2, 1),
                ),
            )

        You can combine this with the `vertex_alignments` parameter

        .. testcode:: multi-building-block-cof-construction

            cof2 = stk.ConstructedMolecule(
                topology_graph=stk.cof.Honeycomb(
                    building_blocks={
                        bb1: (2, 4, 7, 8, 9, 12, 13, 14, 17, 18, 19),
                        bb2: 3,
                        bb3: (0, 10, 11, 15, 16),
                        bb4: (1, 5, 6),
                    },
                    lattice_size=(2, 2, 1),
                    vertex_alignments={0: 2, 1: 1, 2: 1},
                ),
            )

    """

    _allowed_degrees: typing.ClassVar[dict[int, int]]
    _vertex_prototypes: typing.ClassVar[tuple[_CofVertex, ...]]
    _edge_prototypes: typing.ClassVar[tuple[Edge, ...]]
    _lattice_constants: typing.ClassVar[
        tuple[np.ndarray, np.ndarray, np.ndarray]
    ]

    def __init_subclass__(cls, **kwargs):
        vertex_degrees = Counter(
            vertex_id
            for edge in cls._edge_prototypes
            for vertex_id in edge.get_vertex_ids()
        )
        cls._allowed_degrees = set(vertex_degrees.values())

    def __init__(
        self,
        building_blocks: (
            abc.Iterable[BuildingBlock] | dict[BuildingBlock, tuple[int, ...]]
        ),
        lattice_size: tuple[int, int, int],
        periodic: bool = False,
        vertex_alignments: dict[int, int] | None = None,
        reaction_factory: ReactionFactory = GenericReactionFactory(),
        num_processes: int = 1,
        optimizer: Optimizer = NullOptimizer(),
        scale_multiplier: float = 1.0,
    ) -> None:
        """
        Initialize a :class:`.Cof` instance.

        Parameters:

            building_blocks:
                Can be a :class:`tuple` of :class:`.BuildingBlock`
                instances, which should be placed on the topology graph.

                Can also be a :class:`dict` which maps the
                :class:`.BuildingBlock` instances to the ids of the
                vertices it should be placed on. A :class:`dict` is
                required when there are multiple building blocks with the
                same number of functional groups, because in this case
                the desired placement is ambiguous.

            lattice_size:
                The size of the lattice in the x, y and z directions.

            periodic:
                Toggle the construction of a periodic molecule. If
                ``True``, periodic bonds will be made across the edges of
                the lattice.

            vertex_alignments:
                A mapping from the id of a :class:`.Vertex`
                to an :class:`.Edge` connected to it.
                The :class:`.Edge` is used to align the first
                :class:`.FunctionalGroup` of a :class:`.BuildingBlock`
                placed on that vertex. Only vertices which need to have
                their default edge changed need to be present in the
                :class:`dict`. If ``None`` then the default edge is used
                for each vertex. Changing which :class:`.Edge` is used will
                mean that the topology graph represents different
                structural isomers. The edge is referred to by a number
                between ``0`` (inclusive) and the number of edges the
                vertex is connected to (exclusive).

            reaction_factory:
                The reaction factory to use for creating bonds between
                building blocks.

            num_processes:
                The number of parallel processes to create during
                :meth:`construct`.

            optimizer:
                Used to optimize the structure of the constructed
                molecule.

            scale_multiplier:
                Scales the positions of the vertices.

        Raises:

            :class:`AssertionError`
                If the any building block does not have a
                valid number of functional groups.

            :class:`ValueError`
                If the there are multiple building blocks with the
                same number of functional_groups in `building_blocks`,
                and they are not explicitly assigned to vertices. The
                desired placement of building blocks is ambiguous in
                this case.

            :class:`~.cof.UnoccupiedVertexError`
                If a vertex of the COF topology graph does not have a
                building block placed on it.

            :class:`~.cof.OverlyOccupiedVertexError`
                If a vertex of the COF topology graph has more than one
                building block placed on it.

        """

        self._vertex_alignments = (
            dict(vertex_alignments) if vertex_alignments is not None else {}
        )
        self._lattice_size = lattice_size
        self._periodic = periodic

        lattice = self._get_lattice(self._vertex_alignments)
        edges = self._get_edges(lattice)
        vertices = self._get_vertices(lattice)

        if isinstance(building_blocks, dict):
            for building_block in building_blocks:
                assert (
                    building_block.get_num_functional_groups()
                    in self._allowed_degrees
                ), (
                    "The number of functional groups in "
                    f"{building_block} needs to be one of "
                    f"{tuple(self._allowed_degrees)}, but is "
                    "currently "
                    f"{building_block.get_num_functional_groups()}."
                )
            get_vertex = partial(getitem, vertices)
            building_block_vertices = {
                building_block: tuple(
                    map(
                        get_vertex,
                        # Account for the fact that a building block can
                        # be mapped to a single int.
                        (ids,) if isinstance(ids, int) else ids,
                    )
                )
                for building_block, ids in building_blocks.items()
            }
        else:
            building_block_vertices = self._get_building_block_vertices(  # type: ignore[assignment]
                building_blocks=building_blocks,
                vertices=vertices,
                edges=edges,
            )

        building_block_vertices = self._with_unaligning_vertices(
            building_block_vertices=building_block_vertices,
        )

        self._check_building_block_vertices(
            num_vertices=(
                np.prod(lattice_size) * len(self._vertex_prototypes)
            ),
            building_block_vertices=building_block_vertices,
        )

        super().__init__(
            building_block_vertices=typing.cast(
                dict[BuildingBlock, abc.Sequence[Vertex]],
                building_block_vertices,
            ),
            edges=edges,
            reaction_factory=reaction_factory,
            construction_stages=(),
            num_processes=num_processes,
            optimizer=optimizer,
            edge_groups=self._get_edge_groups(edges),
            scale_multiplier=scale_multiplier,
        )

    @staticmethod
    def _with_unaligning_vertices(building_block_vertices):
        clone = dict(building_block_vertices)
        for building_block, vertices in clone.items():
            # Building blocks with 1 placer, cannot be aligned and
            # must therefore use an UnaligningVertex.
            if building_block.get_num_placers() == 1:
                clone[building_block] = tuple(
                    UnaligningVertex(
                        id=v.get_id(),
                        position=v.get_position(),
                        aligner_edge=v.get_aligner_edge(),
                        cell=v.get_cell(),
                    )
                    for v in vertices
                )

        return clone

    @classmethod
    def _check_building_block_vertices(
        cls,
        num_vertices,
        building_block_vertices,
    ):
        unassigned_ids = set(range(num_vertices))
        assigned_ids = set()
        vertices = (
            vertex
            for vertices_ in building_block_vertices.values()
            for vertex in vertices_
        )
        for vertex in vertices:
            if vertex.get_id() in assigned_ids:
                raise OverlyOccupiedVertexError(
                    f"Vertex {vertex.get_id()} has multiple building "
                    "blocks placed on it."
                )
            assigned_ids.add(vertex.get_id())
            unassigned_ids.remove(vertex.get_id())

        if unassigned_ids:
            raise UnoccupiedVertexError(
                f"The following vertices are unoccupied {unassigned_ids}."
            )

    def clone(self):
        clone = super().clone()
        clone._vertex_alignments = dict(self._vertex_alignments)
        clone._lattice_size = self._lattice_size
        clone._periodic = self._periodic
        return clone

    def _get_edge_groups(
        self,
        edges: abc.Iterable[Edge],
    ) -> tuple[EdgeGroup, ...] | None:
        """
        Get the edge groups for the COF.

        Parameters:

            edges:
                The edges which are to be placed into edge groups.

        Returns:

            The edge groups. These are returned when the lattice is
            not periodic, which means that some edges should not be
            part of an edge group, in order to prevent bond formation
            across the edges of the lattice.

            None:
                If an edge group is to be created for every single edge.

        """

        if self._periodic:
            return None

        return tuple(
            EdgeGroup((edge,)) for edge in edges if not edge.is_periodic()
        )

    def _get_vertices(
        self,
        lattice: list[list[list[dict[int, _CofVertex]]]],
    ) -> list[_CofVertex]:
        """
        Get the vertices in the `lattice`.

        Parameters:

            lattice:
                A nested list which can be in the form
                ``lattice[x][y][z][vertex_id]`` which returns a vertex
                in the (x, y, z) cell of the lattice.

        Returns:

            All the vertices extracted from `lattice`.

        """

        xdim, ydim, zdim = self._lattice_size
        vertices = []
        for x, y, z in it.product(
            range(xdim),
            range(ydim),
            range(zdim),
        ):
            for vertex in lattice[x][y][z].values():
                vertices.append(vertex)
        return sorted(vertices, key=lambda vertex: vertex.get_id())

    def _get_lattice(
        self,
        vertex_alignments: dict[int, int],
    ) -> list[list[list[dict[int, _CofVertex]]]]:
        """
        Get the vertices of the topology graph instance.

        Parameters:

            vertex_alignments:
                A mapping from the id of a :class:`.Vertex`
                to an :class:`.Edge` connected to it.
                The :class:`.Edge` is used to align the first
                :class:`.FunctionalGroup` of a :class:`.BuildingBlock`
                placed on that vertex. Only vertices which need to have
                their default edge changed need to be present in the
                :class:`dict`. If ``None`` then the default edge is used
                for each vertex. Changing which :class:`.Edge` is used will
                mean that the topology graph represents different
                structural isomers. The edge is referred to by a number
                between ``0`` (inclusive) and the number of edges the
                vertex is connected to (exclusive).

        Returns:

            A nested :class:`list` which can be indexed as
            ``vertices[x][y][z]``, which will return a :class:`dict`
            for the unit cell at (x, y, z). The :class:`dict` maps
            the vertices in :attr:`_vertex_prototypes` to its clone for
            that unit cell.

        """

        xdim, ydim, zdim = (range(dim) for dim in self._lattice_size)
        # vertex_clones is indexed as vertex_clones[x][y][z]
        lattice: list[list[list[dict[int, _CofVertex]]]] = [
            [[{} for _ in zdim] for _ in ydim] for _ in xdim
        ]
        # Make a clone of each vertex for each unit cell.
        cells = it.product(xdim, ydim, zdim)
        vertices = it.product(cells, self._vertex_prototypes)
        for id_, (cell, vertex) in enumerate(vertices):
            x, y, z = cell
            shift = sum(
                axis * dim for axis, dim in zip(cell, self._lattice_constants)
            )
            lattice[x][y][z][vertex.get_id()] = vertex.__class__(
                id=id_,
                position=vertex.get_position() + shift,
                aligner_edge=vertex_alignments.get(id_, 0),
                cell=cell,
            )
        return lattice

    def _get_edges(self, lattice: list) -> tuple[Edge, ...]:
        """
        Create the edges of the topology graph instance.

        Parameters:

            lattice:
                A nested :class:`list` which can be indexed as
                ``vertices[x][y][z]``, which will return a :class:`dict`
                for the unit cell at (x, y, z). The :class:`dict` maps
                the vertices in :attr:`_vertex_prototypes` to the clones
                for that unit cell.

        Returns:

            The edges of the topology graph instance.

        """

        edge_clones = []
        # Make a clone for each edge for each unit cell.
        xdim, ydim, zdim = (range(dim) for dim in self._lattice_size)
        cells = it.product(xdim, ydim, zdim)
        edges = it.product(cells, self._edge_prototypes)
        for id_, (cell, edge) in enumerate(edges):
            x, y, z = cell
            # The cell in which the second vertex of the edge is found.
            periodic_cell = np.array(cell) + edge.get_periodicity()
            # Wrap around periodic cells, ie those that are less than 0
            # or greater than the lattice size along any dimension.
            dims = zip(periodic_cell, self._lattice_size)
            x2, y2, z2 = np.array(
                [(dim + max_dim) % max_dim for dim, max_dim in dims]
            )
            # The edge is not periodic if periodic_cell did not
            # have to wrap around.
            dims = zip(periodic_cell, self._lattice_size)
            edge_is_not_periodic = all(
                dim >= 0 and dim < max_dim for dim, max_dim in dims
            )
            edge_clones.append(
                CofEdge(
                    parent_id=edge.get_id(),
                    id=id_,
                    vertex1=lattice[x][y][z][edge.get_vertex1_id()],
                    vertex2=lattice[x2][y2][z2][edge.get_vertex2_id()],
                    periodicity=(
                        (0, 0, 0)
                        if edge_is_not_periodic
                        else edge.get_periodicity()
                    ),
                )
            )

        return tuple(edge_clones)

    @classmethod
    def _get_building_block_vertices(
        cls,
        building_blocks: abc.Iterable[BuildingBlock],
        vertices: abc.Iterable[_CofVertex],
        edges: abc.Iterable[Edge],
    ) -> dict[BuildingBlock, abc.Sequence[_CofVertex]]:
        """
        Map building blocks to the vertices of the graph.

        Parameters:

            building_blocks:
                The building blocks which need to be mapped to `vertices`.

            vertices:
                The vertices which need to have a building block map to
                them.

            edges:
                The edges of the graph.

        Returns:

            Maps each building block in `building_blocks` to a
            :class:`list` of the :class:`.Vertex` instances it should
            be placed on.

        Raises:

            :class:`AssertionError`
                If the any building block does not have a
                valid number of functional groups.

            :class:`ValueError`
                If there are multiple building blocks with the same number
                of functional groups.

        """

        building_blocks_by_degree = {}
        for building_block in building_blocks:
            num_fgs = building_block.get_num_functional_groups()
            assert num_fgs in cls._allowed_degrees, (
                "The number of functional groups in "
                f"{building_block} needs to be one of "
                f"{tuple(cls._allowed_degrees)}, but is "
                "currently "
                f"{building_block.get_num_functional_groups()}."
            )
            if num_fgs in building_blocks_by_degree:
                raise ValueError(
                    "If there are multiple building blocks with the "
                    "same number of functional groups, "
                    "building_block_vertices must be set explicitly."
                )
            building_blocks_by_degree[num_fgs] = building_block

        vertex_degrees = Counter(
            vertex_id for edge in edges for vertex_id in edge.get_vertex_ids()
        )

        building_block_vertices = {}  # type: ignore[var-annotated]
        for vertex in vertices:
            vertex_degree = vertex_degrees[vertex.get_id()]
            building_block = building_blocks_by_degree[vertex_degree]
            building_block_vertices[building_block] = (
                building_block_vertices.get(building_block, [])
            )
            building_block_vertices[building_block].append(vertex)
        return building_block_vertices

    def _get_lattice_constants(self) -> abc.Iterator[np.ndarray]:
        yield from self._lattice_constants

    def construct(self) -> ConstructionResult:
        """
        Construct a :class:`.ConstructedMolecule`.

        Returns:

            The data describing the :class:`.ConstructedMolecule`.

        """

        return super().construct()

    def _get_construction_result(self, state):
        return PeriodicConstructionResult(state, self._lattice_size)

    @staticmethod
    def _get_scale(
        building_block_vertices: dict[BuildingBlock, abc.Sequence[Vertex]],
        scale_multiplier: float,
    ) -> float:
        return (
            scale_multiplier
            * 5
            * max(bb.get_maximum_diameter() for bb in building_block_vertices)
        )

    def __repr__(self):
        x, y, z = self._lattice_size
        periodic = ", periodic=True" if self._periodic else ""
        vertex_alignments = (
            f", vertex_alignments={self._vertex_alignments}"
            if self._vertex_alignments
            else ""
        )
        return (
            f"cof.{self.__class__.__name__}("
            f"({x}, {y}, {z})"
            f"{vertex_alignments}"
            f"{periodic})"
        )



import pytest

import stk

from ....case_data import CaseData


def _get_fake_complex() -> stk.BuildingBlock:
    return stk.BuildingBlock(
        smiles="BrCc1cc(CBr)cc(CBr)c1",
        functional_groups=[stk.BromoFactory()],
    )


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cage.M4L6Tetrahedron(
                    building_blocks={_get_fake_complex(): range(4)},
                ),
            ),
            smiles=(
                "[H]C1=C2C([H])=C3C([H])=C1C([H])([H])C([H])([H])C1=C("
                "[H])C4=C([H])C(=C1[H])C([H])([H])C([H])([H])C1=C([H])"
                "C(=C([H])C(=C1[H])C([H])([H])C3([H])[H])C([H])([H])C("
                "[H])([H])C1=C([H])C(=C([H])C(=C1[H])C([H])([H])C2([H]"
                ")[H])C([H])([H])C4([H])[H]"
            ),
            name=name,
        ),
    ),
)
def metal_cage_m4l6_tetrahedron(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



import stk

from .utilities import is_equivalent_atom


def test_repr(atom):
    """
    Test :meth:`.Atom.__repr__`.

    Parameters
    ----------
    atom : :class:`.Atom`
        The atom, whose representation should be tested.

    Returns
    -------
    None : :class:`NoneType`

    """

    other = eval(repr(atom), dict(stk.__dict__))
    is_equivalent_atom(other, atom)



"""
MAE Updating Utilities
======================

"""

from stk._internal.utilities.utilities import mol_from_mae_file


def _with_structure_from_mae(self, path):
    """
    Change structure to match an ``.mae`` file.

    Parameters
    ----------
    path : :class:`str`
        The full path of the ``.mae`` file from which the structure
        should be updated.

    Returns
    -------
    :class:`.Molecule`
        The molecule.

    """

    molecule = mol_from_mae_file(path)
    return self._with_position_matrix(
        position_matrix=molecule.GetConformer().GetPositions()
    )



import numpy as np


def test_clone(periodic_case):
    """
    Test :meth:`.PeriodicInfo.clone`.

    Parameters
    ----------
    periodic_case : :class:`.CaseData`
        The periodic case to test.

    Returns
    -------
    None : :class:`NoneType`

    """

    original = periodic_case.periodic_info
    clone = periodic_case.periodic_info.clone()

    assert np.all(
        np.equal(
            np.array(clone.get_cell_matrix()),
            np.array(original.get_cell_matrix()),
        )
    )
    assert np.all(
        np.equal(
            clone.get_vector_1(),
            original.get_vector_1(),
        )
    )
    assert np.all(
        np.equal(
            clone.get_vector_2(),
            original.get_vector_2(),
        )
    )
    assert np.all(
        np.equal(
            clone.get_vector_3(),
            original.get_vector_3(),
        )
    )
    assert clone.get_a() == original.get_a()
    assert clone.get_b() == original.get_b()
    assert clone.get_c() == original.get_c()
    assert clone.get_alpha() == original.get_alpha()
    assert clone.get_beta() == original.get_beta()
    assert clone.get_gamma() == original.get_gamma()



from stk._internal.molecule import Molecule

from .molecule import MoleculeKeyMaker
from .utilities import get_smiles


class Smiles(MoleculeKeyMaker):
    """
    Used to get the SMILES of molecules.

    .. warning::

        SMILES strings generated with ``stk`` version ``v2025.07.17.0`` or sooner
        (using ``rdkit`` version ``2024.9.1`` or sooner) will be different than newer
        versions due to a change in handling the valence of organic atoms bound to
        metals. Details can be found in the ``rdkit`` release notes `2025_03_1`_.
        No changes occur in ``stk`` construction. An example change in SMILES:
        ``CCCO->[Fe+2]`` becomes ``CCC[OH]->[Fe+2]``.

    .. _`2025_03_1`: https://github.com/rdkit/rdkit/releases/tag/Release_2025_03_1

    Examples:

        *Adding SMILES to a Molecule's JSON*

        You want to use the isomeric, canonical SMILES from RDKit as
        part of a JSON representation of a molecule

        .. testcode:: adding-smiles-to-a-molecules-json

            import stk

            jsonizer = stk.MoleculeJsonizer(
                key_makers=(stk.Smiles(), ),
            )
            # Get the JSON representation, including an SMILES.
            json = jsonizer.to_json(stk.BuildingBlock('NCCN'))

        .. testcode:: adding-smiles-to-a-molecules-json
            :hide:

            assert json['molecule']['SMILES'] == 'NCCN'
            assert json['matrix']['SMILES'] == 'NCCN'
    """

    def __init__(self) -> None:
        return

    def get_key_name(self) -> str:
        return "SMILES"

    def get_key(self, molecule: Molecule) -> str:
        return get_smiles(molecule)

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return "Smiles()"



from tests.utilities import is_equivalent_constructed_molecule


def test_database(case_data):
    """
    Test a :class:`.MoleculeDatabase`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the database to test and the molecule to
        place into the database.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_database(
        database=case_data.database,
        molecule=case_data.molecule,
        key=case_data.key,
    )


def _test_database(database, molecule, key):
    """
    Test a molecule or constructed molecule database.

    Parameters
    ----------
    database : :class:`.MoleculeCache`
        The database to test.

    molecule : :class:`.Molecule`
        The molecule to put and get from the `database`.

    key : :class:`object`
        The key used to retrieve `molecule` from the database.

    Returns
    -------
    None : :class:`NoneType`

    """

    database.put(molecule)
    retrieved = database.get(key)
    is_equivalent_constructed_molecule(
        molecule.with_canonical_atom_ordering(),
        retrieved.with_canonical_atom_ordering(),
    )



"""
Atom Batch
==========

"""

from __future__ import annotations

from typing import Iterable

from stk._internal.atom import Atom
from stk._internal.atom_info import AtomInfo
from stk._internal.building_block import BuildingBlock


class _AtomBatch:
    """
    A batch of atoms.

    """

    __slots__ = ["_atoms", "_atom_infos", "_id_map"]

    _atoms: tuple[Atom, ...]
    _atom_infos: tuple[AtomInfo, ...]
    _id_map: dict[int, int]

    def __init__(
        self,
        atoms: Iterable[Atom],
        num_atoms: int,
        building_block: BuildingBlock,
        building_block_id: int,
    ) -> None:
        """
        Initialize an :class:`._AtomBatch` instance.

        Parameters:

            atoms:
                The atoms, which should be added to the batch.

            num_atoms:
                The number of atoms in the molecule being constructed,
                before atoms in this batch are taken into account.

            building_block:
                The building block from which the atoms originate.

            building_block_id:
                An id, unique to that building block and placement.

        """

        _atoms = []
        atom_infos = []
        self._id_map = {}

        for id_, atom in enumerate(atoms, num_atoms):
            _atoms.append(atom.with_id(id_))
            self._id_map[atom.get_id()] = _atoms[-1].get_id()
            atom_infos.append(
                AtomInfo(
                    atom=_atoms[-1],
                    building_block_atom=atom,
                    building_block=building_block,
                    building_block_id=building_block_id,
                )
            )

        self._atoms = tuple(_atoms)
        self._atom_infos = tuple(atom_infos)

    def get_atoms(self) -> Iterable[Atom]:
        """
        Yield the atoms in the batch.

        Yields:

            An atom.

        """

        yield from self._atoms

    def get_atom_infos(self) -> Iterable[AtomInfo]:
        """
        Yield info about the atoms in the batch.

        Yields:

            Info about an atom in the batch.

        """

        yield from self._atom_infos

    def get_id_map(self) -> dict[int, int]:
        """
        Get a mapping from the old atom id to the new atom.

        Returns:

            Maps the id of an atom provided to the initializer, to
            the id of the new, corresponding, atom held by the batch.

        """

        return dict(self._id_map)



"""
Bond Batch
==========


"""

from stk._internal.bond_info import BondInfo


class _BondBatch:
    """
    A batch of bonds.

    """

    __slots__ = ["_bonds", "_bond_infos"]

    def __init__(self, bonds, atom_map):
        """
        Initialize a :class:`.BondBatch` instance.

        Parameters
        ----------
        bonds : :class:`iterable` of :class:`.Bond`
            The bonds, which should be added to the batch.

        atom_map : :class:`dict`
            Maps the ids of atoms held by `bonds`, to the new atoms,
            which the bonds in the :class:`.BondBatchData` instance
            should hold.

        """

        self._bonds = _bonds = []
        self._bond_infos = bond_infos = []

        for bond in bonds:
            _bonds.append(bond.with_atoms(atom_map))
            bond_infos.append(BondInfo(_bonds[-1], None, None))

    def get_bonds(self):
        """
        Yield the bonds in the batch.

        Yields
        ------
        :class:`.Bond`
            A bond in the batch.

        """

        yield from self._bonds

    def get_bond_infos(self):
        """
        Yield info about the bonds in the batch.

        Yields
        ------
        :class:`.BondInfo`
            Info about a bond in the batch.

        """

        yield from self._bond_infos



from stk._internal.topology_graphs.metal_complex.bidentate_square_planar import (  # noqa
    BidentateSquarePlanar,
)
from stk._internal.topology_graphs.metal_complex.cis_protected_square_planar import (  # noqa
    CisProtectedSquarePlanar,
)
from stk._internal.topology_graphs.metal_complex.metal_complex import (
    MetalComplex,
)
from stk._internal.topology_graphs.metal_complex.octahedral import Octahedral
from stk._internal.topology_graphs.metal_complex.octahedral_delta import (
    OctahedralDelta,
)
from stk._internal.topology_graphs.metal_complex.octahedral_lambda import (
    OctahedralLambda,
)
from stk._internal.topology_graphs.metal_complex.paddlewheel import Paddlewheel
from stk._internal.topology_graphs.metal_complex.porphyrin import Porphyrin
from stk._internal.topology_graphs.metal_complex.square_planar import (
    SquarePlanar,
)
from stk._internal.topology_graphs.metal_complex.vertices import (
    BiDentateLigandVertex,
    MonoDentateLigandVertex,
    MetalVertex,
)

__all__ = [
    "Porphyrin",
    "Octahedral",
    "OctahedralDelta",
    "OctahedralLambda",
    "MetalComplex",
    "Paddlewheel",
    "SquarePlanar",
    "BidentateSquarePlanar",
    "CisProtectedSquarePlanar",
    "BiDentateLigandVertex",
    "MonoDentateLigandVertex",
    "MetalVertex",
]



"""
Alcohol Factory
===============

"""

from stk._internal.functional_group_factories.functional_group_factory import (
    FunctionalGroupFactory,
)
from stk._internal.functional_group_factories.utilities import get_atom_ids
from stk._internal.functional_groups.alcohol import Alcohol


class AlcoholFactory(FunctionalGroupFactory):
    """
    Creates :class:`.Alcohol` instances.

    Creates functional groups from substructures, which match the
    ``[*][O][H]`` functional group string.

    Examples
    --------
    *Creating Functional Groups with the Factory*

    You want to create a building block which has :class:`.Alcohol`
    functional groups. You want the oxygen atom in those functional
    groups to be the bonder atom, and the hydrogen atom to be the
    deleter atom.

    .. testcode:: creating-functional-groups-with-the-factory

        import stk

        building_block = stk.BuildingBlock(
            smiles='OCCCO',
            functional_groups=(stk.AlcoholFactory(), ),
        )

    .. testcode:: creating-functional-groups-with-the-factory
        :hide:

        assert all(
            isinstance(functional_group, stk.Alcohol)
            for functional_group
            in building_block.get_functional_groups()
        )
        assert building_block.get_num_functional_groups() == 2

    *Changing the Bonder and Deleter Atoms*

    You want to create a building block which has :class:`.Alcohol`
    functional groups. You want the OH group to be
    treated as a leaving group. This means the non-hydrogen atom bonded
    to oxygen is the *bonder* atom and both the oxygen and hydrogen
    atoms are *deleter* atoms.

    .. testcode:: changing-the-bonder-and-deleter-atoms

        import stk

        alcohol_factory = stk.AlcoholFactory(
            # The index of the non-hydrogen atom connected to oxygen
            # is 0 in the functional group string (see docstring).
            bonders=(0, ),
            # The indices of the oxygen and hydrogen atoms in the
            # functional group string (see docstring) are
            # 1 and 2, respectively.
            deleters=(1, 2),
        )
        building_block = stk.BuildingBlock(
            smiles='OCCCO',
            functional_groups=(alcohol_factory, ),
        )

    .. testcode:: changing-the-bonder-and-deleter-atoms
        :hide:

        fg1, fg2 = building_block.get_functional_groups()
        assert fg1.get_num_bonders() == 1
        assert sum(1 for _ in fg1.get_deleters()) == 2
        assert fg2.get_num_bonders() == 1
        assert sum(1 for _ in fg2.get_deleters()) == 2

        assert all(
            isinstance(atom, stk.C)
            for functional_group
            in building_block.get_functional_groups()
            for atom
            in functional_group.get_bonders()
        )
        assert all(
            isinstance(atom, (stk.O, stk.H))
            for functional_group
            in building_block.get_functional_groups()
            for atom
            in functional_group.get_deleters()
        )

    See Also
    --------
    :class:`.GenericFunctionalGroup`
        Defines *bonders* and  *deleters*.

    """

    def __init__(self, bonders=(1,), deleters=(2,), placers=None):
        """
        Initialize an :class:`.AlcoholFactory` instance.

        Parameters
        ----------
        bonders : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are *bonder* atoms.

        deleters : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are *deleter* atoms.

        placers : :class:`tuple` of :class:`int`, optional
            The indices of atoms in the functional group string, which
            are *placer* atoms. If ``None``, `bonders` will be used.

        """

        self._bonders = bonders
        self._deleters = deleters
        self._placers = bonders if placers is None else placers

    def get_functional_groups(self, molecule):
        for atom_ids in get_atom_ids("[*][O][H]", molecule):
            atoms = tuple(molecule.get_atoms(atom_ids))
            yield Alcohol(
                oxygen=atoms[1],
                hydrogen=atoms[2],
                atom=atoms[0],
                bonders=tuple(atoms[i] for i in self._bonders),
                deleters=tuple(atoms[i] for i in self._deleters),
                placers=tuple(atoms[i] for i in self._placers),
            )



"""
M6L2L3 Prism
============

"""

import numpy as np

from stk._internal.topology_graphs.edge import Edge

from .cage import Cage
from .vertices import NonLinearVertex


class M6L2L3Prism(Cage):
    """
    Represents a cage topology graph.

    Unoptimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        iron_atom = stk.BuildingBlock(
            smiles='[Fe+2]',
            functional_groups=(
                stk.SingleAtom(stk.Fe(0, charge=2))
                for i in range(6)
            ),
            position_matrix=[[0, 0, 0]],
        )

        bb2 = stk.BuildingBlock(
            smiles='C1=NC(C=NBr)=CC=C1',
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#35]',
                    bonders=(1, ),
                    deleters=(),
                ),
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )

        complex = stk.ConstructedMolecule(
            topology_graph=stk.metal_complex.OctahedralDelta(
                metals=iron_atom,
                ligands=bb2,
                optimizer=stk.MCHammer(),
            ),
        )

        # Assign Bromo functional groups to the metal complex.
        iron_oct_delta = stk.BuildingBlock.init_from_molecule(
            molecule=complex,
            functional_groups=[stk.BromoFactory()],
        )

        # Define building blocks.
        bb3 = stk.BuildingBlock(
            smiles=(
                'C1=C(C=C(C=C1Br)Br)Br'
            ),
            functional_groups=[stk.BromoFactory()],
        )
        bb4 = stk.BuildingBlock(
            smiles=(
                'C1=CC(=CC=C1C2=CC(=C(C=C2C3=CC=C(C=C3)Br)C4=CC=C(C=C'
                '4)Br)C5=CC=C(C=C5)Br)Br'
            ),
            functional_groups=[stk.BromoFactory()],
        )

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.M6L2L3Prism(
                building_blocks={
                    iron_oct_delta: range(6),
                    bb3: (6, 7),
                    bb4: range(8, 11),
                },
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in cage.get_bonds()
            ),
        )

    :class:`.Collapser` optimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        iron_atom = stk.BuildingBlock(
            smiles='[Fe+2]',
            functional_groups=(
                stk.SingleAtom(stk.Fe(0, charge=2))
                for i in range(6)
            ),
            position_matrix=[[0, 0, 0]],
        )

        bb2 = stk.BuildingBlock(
            smiles='C1=NC(C=NBr)=CC=C1',
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#35]',
                    bonders=(1, ),
                    deleters=(),
                ),
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )

        complex = stk.ConstructedMolecule(
            topology_graph=stk.metal_complex.OctahedralDelta(
                metals=iron_atom,
                ligands=bb2,
                optimizer=stk.MCHammer(),
            ),
        )

        # Assign Bromo functional groups to the metal complex.
        iron_oct_delta = stk.BuildingBlock.init_from_molecule(
            molecule=complex,
            functional_groups=[stk.BromoFactory()],
        )

        # Define building blocks.
        bb3 = stk.BuildingBlock(
            smiles=(
                'C1=C(C=C(C=C1Br)Br)Br'
            ),
            functional_groups=[stk.BromoFactory()],
        )
        bb4 = stk.BuildingBlock(
            smiles=(
                'C1=CC(=CC=C1C2=CC(=C(C=C2C3=CC=C(C=C3)Br)C4=CC=C(C=C'
                '4)Br)C5=CC=C(C=C5)Br)Br'
            ),
            functional_groups=[stk.BromoFactory()],
        )

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.M6L2L3Prism(
                building_blocks={
                    iron_oct_delta: range(6),
                    bb3: (6, 7),
                    bb4: range(8, 11),
                },
                optimizer=stk.Collapser(),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in cage.get_bonds()
            ),
        )

    Metal building blocks with three functional groups are
    required for this topology.

    Ligand building blocks with three and four functional groups are
    required for ligand type A and B, respectively, on this topology.

    When using a :class:`dict` for the `building_blocks` parameter,
    as in :ref:`cage-topology-graph-examples`:
    *Multi-Building Block Cage Construction*, a
    :class:`.BuildingBlock`, with the following number of functional
    groups, needs to be assigned to each of the following vertex ids:

        | 3-functional groups (metal): 0 to 5
        | 3-functional groups (ligand A): 6, 7
        | 4-functional groups (ligand B): 8 to 10

    See :class:`.Cage` for more details and examples.

    """

    _non_linears = (
        NonLinearVertex(0, np.array([-1, -1 / np.sqrt(3), 1])),
        NonLinearVertex(1, np.array([1, -1 / np.sqrt(3), 1])),
        NonLinearVertex(2, np.array([0, 2 / np.sqrt(3), 1])),
        NonLinearVertex(3, np.array([-1, -1 / np.sqrt(3), -1])),
        NonLinearVertex(4, np.array([1, -1 / np.sqrt(3), -1])),
        NonLinearVertex(5, np.array([0, 2 / np.sqrt(3), -1])),
    )

    _vertex_prototypes = (
        *_non_linears,
        NonLinearVertex.init_at_center(
            id=6,
            vertices=(
                _non_linears[0],
                _non_linears[1],
                _non_linears[2],
            ),
        ),
        NonLinearVertex.init_at_center(
            id=7,
            vertices=(
                _non_linears[3],
                _non_linears[4],
                _non_linears[5],
            ),
        ),
        NonLinearVertex.init_at_center(
            id=8,
            vertices=(
                _non_linears[0],
                _non_linears[1],
                _non_linears[3],
                _non_linears[4],
            ),
        ),
        NonLinearVertex.init_at_center(
            id=9,
            vertices=(
                _non_linears[1],
                _non_linears[2],
                _non_linears[4],
                _non_linears[5],
            ),
        ),
        NonLinearVertex.init_at_center(
            id=10,
            vertices=(
                _non_linears[2],
                _non_linears[0],
                _non_linears[5],
                _non_linears[3],
            ),
        ),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[6]),
        Edge(1, _vertex_prototypes[0], _vertex_prototypes[8]),
        Edge(2, _vertex_prototypes[0], _vertex_prototypes[10]),
        Edge(3, _vertex_prototypes[1], _vertex_prototypes[6]),
        Edge(4, _vertex_prototypes[1], _vertex_prototypes[8]),
        Edge(5, _vertex_prototypes[1], _vertex_prototypes[9]),
        Edge(6, _vertex_prototypes[2], _vertex_prototypes[6]),
        Edge(7, _vertex_prototypes[2], _vertex_prototypes[9]),
        Edge(8, _vertex_prototypes[2], _vertex_prototypes[10]),
        Edge(9, _vertex_prototypes[3], _vertex_prototypes[7]),
        Edge(10, _vertex_prototypes[3], _vertex_prototypes[8]),
        Edge(11, _vertex_prototypes[3], _vertex_prototypes[10]),
        Edge(12, _vertex_prototypes[4], _vertex_prototypes[7]),
        Edge(13, _vertex_prototypes[4], _vertex_prototypes[8]),
        Edge(14, _vertex_prototypes[4], _vertex_prototypes[9]),
        Edge(15, _vertex_prototypes[5], _vertex_prototypes[7]),
        Edge(16, _vertex_prototypes[5], _vertex_prototypes[9]),
        Edge(17, _vertex_prototypes[5], _vertex_prototypes[10]),
    )

    _num_windows = 4
    _num_window_types = 1



def test_is_periodic(case_data):
    """
    Test :meth:`.Bond.is_periodic`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the bond to test and its true periodicity.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_is_periodic(case_data.bond, case_data.periodicity)


def _test_is_periodic(bond, periodicity):
    """
    Test :meth:`.Bond.is_periodic`.

    Parameters
    ----------
    bond : :class:`.Bond`
        The bond to test.

    periodicity : :class:`tuple` of :class:`int`
        The true periodicity of `bond`.

    Returns
    -------
    None : :class:`NoneType`

    """

    return bond.is_periodic() == any(p != 0 for p in periodicity)



import numpy as np


def get_centroid(position_matrix, atom_ids):
    # In this case, Molecule.get_centroid should raise the error, so
    # this should should just fail silently, and the explicit failure
    # of Molecule.get_centroid will be caught.
    if len(atom_ids) == 0:
        return

    return np.sum(position_matrix[atom_ids, :], axis=0) / len(atom_ids)


def get_direction(position_matrix, atom_ids):
    # In this case, Molecule.get_direction should raise the error, so
    # this should should just fail silently, and the explicit failure
    # of Molecule.get_direction will be caught.
    if len(atom_ids) == 0:
        return

    atom_positions = position_matrix[atom_ids, :]
    centered_positions = atom_positions - atom_positions.mean(axis=0)
    return np.around(np.linalg.svd(centered_positions)[-1][0], 14)



"""
Carboxylic Acid
===============

"""

from .generic_functional_group import GenericFunctionalGroup


class CarboxylicAcid(GenericFunctionalGroup):
    """
    Represents a carboxylic acid functional group.

    The structure of the functional group is given by the pseudo-SMILES
    ``[atom][carbon](=[oxygen1])[oxygen2][hydrogen]``.

    """

    def __init__(
        self,
        carbon,
        oxygen1,
        oxygen2,
        hydrogen,
        atom,
        bonders,
        deleters,
        placers=None,
    ):
        """
        Initialize a :class:`.CarboxylicAcid` instance.

        Parameters
        ----------
        carbon : :class:`.C`
            The ``[carbon]`` atom.

        oxygen1 : :class:`.O`
            The ``[oxygen1]`` atom.

        oxygen2 : :class:`.O`
            The ``[oxygen2]`` atom.

        hydrogen : :class:`.H`
            The ``[hydrogen]`` atom.

        atom : :class:`.Atom`
            The ``[atom]`` atom.

        bonders : :class:`tuple` of :class:`.Atom`
            The bonder atoms.

        deleters : :class:`tuple` of :class:`.Atom`
            The deleter atoms.

        placers : :class:`tuple` of :class:`.Atom`, optional
            The placer atoms. If ``None`` the `bonders` will be used.

        """

        self._carbon = carbon
        self._oxygen1 = oxygen1
        self._oxygen2 = oxygen2
        self._hydrogen = hydrogen
        self._atom = atom
        super().__init__(
            atoms=(carbon, oxygen1, oxygen2, hydrogen, atom),
            bonders=bonders,
            deleters=deleters,
            placers=bonders if placers is None else placers,
        )

    def get_carbon(self):
        """
        Get the ``[carbon]`` atom.

        Returns
        -------
        :class:`.C`
            The ``[carbon]`` atom.

        """

        return self._carbon

    def get_oxygen1(self):
        """
        Get the ``[oxygen1]`` atom.

        Returns
        -------
        :class:`.O`
            The ``[oxygen]`` atom.

        """

        return self._oxygen1

    def get_oxygen2(self):
        """
        Get the ``[oxygen2]`` atom.

        Returns
        -------
        :class:`.O`
            The ``[oxygen2]`` atom.

        """

        return self._oxygen2

    def get_hydrogen(self):
        """
        Get the ``[hydrogen]`` atom.

        Returns
        -------
        :class:``
            The ``[hydrogen]`` atom.

        """

        return self._hydrogen

    def get_atom(self):
        """
        Get the ``[atom]`` atom.

        Returns
        -------
        :class:``
            The ``[atom]`` atom.

        """

        return self._atom

    def clone(self):
        clone = super().clone()
        clone._carbon = self._carbon
        clone._oxygen1 = self._oxygen1
        clone._oxygen2 = self._oxygen2
        clone._hydrogen = self._hydrogen
        clone._atom = self._atom
        return clone

    def with_atoms(self, atom_map):
        clone = super().with_atoms(atom_map)
        clone._carbon = atom_map.get(
            self._carbon.get_id(),
            self._carbon,
        )
        clone._oxygen1 = atom_map.get(
            self._oxygen1.get_id(),
            self._oxygen1,
        )
        clone._oxygen2 = atom_map.get(
            self._oxygen2.get_id(),
            self._oxygen2,
        )
        clone._hydrogen = atom_map.get(
            self._hydrogen.get_id(),
            self._hydrogen,
        )
        clone._atom = atom_map.get(
            self._atom.get_id(),
            self._atom,
        )
        return clone

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"{self._carbon}, {self._oxygen1}, {self._oxygen2}, "
            f"{self._hydrogen}, {self._atom}, "
            f"bonders={self._bonders}, deleters={self._deleters})"
        )



"""
Reaction Result
===============

"""


class ReactionResult:
    """
    The result of a reaction.

    """

    __slots__ = [
        "_new_atoms",
        "_new_bonds",
        "_deleted_atoms",
        "_deleted_bonds",
    ]

    def __init__(
        self,
        new_atoms,
        new_bonds,
        deleted_atoms,
        deleted_bonds,
    ):
        """
        Initialize a :class:`.ReactionResult` instance.

        Parameters
        ----------
        new_atoms : :class:`tuple` of :class:`.NewAtom`
            The new atoms added by the reaction.

        new_bonds : :class:`tuple` of :class:`.Bond`
            The bonds added by the reaction.

        deleted_atoms : :class:`tuple` of :class:`.Atom`
            The atoms deleted by the reaction.

        deleted_bonds : :class:`tuple` of :class:`.Bond`
            The bonds deleted by the reaction.

        """

        self._new_atoms = new_atoms
        self._new_bonds = new_bonds
        self._deleted_atoms = deleted_atoms
        self._deleted_bonds = deleted_bonds

    def get_new_atoms(self):
        """
        Get the new atoms added by the reaction.

        Returns
        -------
        :class:`tuple` of :class:`.NewAtom`
            The new atoms added by the reaction.

        """

        return self._new_atoms

    def get_new_bonds(self):
        """
        Get the new bonds added by the reaction.

        Returns
        -------
        :class:`tuple` of :class:`.Bond`
            The new bonds added by the reaction.

        """
        return self._new_bonds

    def get_deleted_atoms(self):
        """
        Get the atoms deleted by the reaction.

        Returns
        -------
        :class:`tuple` of :class:`.Atom`
            The atoms deleted by the reaction.

        """

        return self._deleted_atoms

    def get_deleted_bonds(self):
        """
        Get the bonds deleted by the reaction.

        Returns
        -------
        :class:`tuple` of :class:`.Bond`
            The bonds deleted by the reaction.

        """

        return self._deleted_bonds



import numpy as np
import pytest

import stk

from ....case_data import CaseData
from ...building_blocks import (
    get_other_linker,
    get_palladium_bi_1,
    get_pd_atom,
)


def _get_palladium_cispbi_sqpl() -> stk.BuildingBlock:
    palladium_cispbi_sqpl = stk.ConstructedMolecule(
        topology_graph=stk.metal_complex.CisProtectedSquarePlanar(
            metals=get_pd_atom(),
            ligands=get_palladium_bi_1(),
        ),
    )
    return stk.BuildingBlock.init_from_molecule(
        molecule=palladium_cispbi_sqpl,
        functional_groups=[
            stk.SmartsFunctionalGroupFactory(
                smarts="[Pd]~[#7]",
                bonders=(0,),
                deleters=(),
                placers=(0, 1),
            ),
        ],
    )


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                stk.cage.M4L4Square(
                    corners=_get_palladium_cispbi_sqpl(),
                    linkers=get_other_linker(),
                    reaction_factory=stk.DativeReactionFactory(
                        stk.GenericReactionFactory(
                            bond_orders={
                                frozenset(
                                    {
                                        stk.GenericFunctionalGroup,
                                        stk.GenericFunctionalGroup,
                                    }
                                ): 9,
                            }
                        )
                    ),
                )
            ),
            smiles=(
                "[H]C1=C2C3=C([H])C([H])=[N](->[Pd+2]4(<-[N]5=C([H])C([H])=C"
                "(C6=C([H])C([H])=[N](->[Pd+2]7(<-[N]8=C([H])C([H])=C(C9=C(["
                "H])C([H])=[N](->[Pd+2]%10(<-[N]%11=C([H])C([H])=C(C%12=C([H"
                "])C([H])=[N](->[Pd+2]%13(<-[N](=C1[H])C([H])=C2[H])<-[N]([H"
                "])([H])C([H])([H])C([H])([H])[N]->%13([H])[H])C([H])=C%12[H"
                "])C([H])=C%11[H])<-[N]([H])([H])C([H])([H])C([H])([H])[N]->"
                "%10([H])[H])C([H])=C9[H])C([H])=C8[H])<-[N]([H])([H])C([H])"
                "([H])C([H])([H])[N]->7([H])[H])C([H])=C6[H])C([H])=C5[H])<-"
                "[N]([H])([H])C([H])([H])C([H])([H])[N]->4([H])[H])C([H])=C3"
                "[H]"
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cage.M4L4Square(
                    corners=_get_palladium_cispbi_sqpl(),
                    linkers=get_other_linker(),
                    reaction_factory=stk.DativeReactionFactory(
                        reaction_factory=stk.GenericReactionFactory(
                            bond_orders={
                                frozenset(
                                    {
                                        stk.GenericFunctionalGroup,
                                        stk.GenericFunctionalGroup,
                                    }
                                ): 9,
                            }
                        )
                    ),
                    optimizer=stk.MCHammer(
                        num_steps=150,
                        random_seed=1000,
                    ),
                ),
            ),
            smiles=(
                "[H]C1=C2C3=C([H])C([H])=[N](->[Pd+2]4(<-[N]5=C([H])C([H])=C("
                "C6=C([H])C([H])=[N](->[Pd+2]7(<-[N]8=C([H])C([H])=C(C9=C([H]"
                ")C([H])=[N](->[Pd+2]%10(<-[N]%11=C([H])C([H])=C(C%12=C([H])C"
                "([H])=[N](->[Pd+2]%13(<-[N](=C1[H])C([H])=C2[H])<-[N]([H])(["
                "H])C([H])([H])C([H])([H])[N]->%13([H])[H])C([H])=C%12[H])C(["
                "H])=C%11[H])<-[N]([H])([H])C([H])([H])C([H])([H])[N]->%10([H"
                "])[H])C([H])=C9[H])C([H])=C8[H])<-[N]([H])([H])C([H])([H])C("
                "[H])([H])[N]->7([H])[H])C([H])=C6[H])C([H])=C5[H])<-[N]([H])"
                "([H])C([H])([H])C([H])([H])[N]->4([H])[H])C([H])=C3[H]"
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cage.M4L4Square(
                    corners=_get_palladium_cispbi_sqpl(),
                    linkers=get_other_linker(),
                    reaction_factory=stk.DativeReactionFactory(
                        reaction_factory=stk.GenericReactionFactory(
                            bond_orders={
                                frozenset(
                                    {
                                        stk.GenericFunctionalGroup,
                                        stk.GenericFunctionalGroup,
                                    }
                                ): 9,
                            }
                        )
                    ),
                    vertex_positions={0: np.array([2, 0, 2])},
                    optimizer=stk.MCHammer(
                        num_steps=150,
                        random_seed=1000,
                    ),
                ),
            ),
            smiles=(
                "[H]C1=C2C3=C([H])C([H])=[N](->[Pd+2]4(<-[N]5=C([H])C([H])=C"
                "(C6=C([H])C([H])=[N](->[Pd+2]7(<-[N]8=C([H])C([H])=C(C9=C(["
                "H])C([H])=[N](->[Pd+2]%10(<-[N]%11=C([H])C([H])=C(C%12=C([H"
                "])C([H])=[N](->[Pd+2]%13(<-[N](=C1[H])C([H])=C2[H])<-[N]([H"
                "])([H])C([H])([H])C([H])([H])[N]->%13([H])[H])C([H])=C%12[H"
                "])C([H])=C%11[H])<-[N]([H])([H])C([H])([H])C([H])([H])[N]->"
                "%10([H])[H])C([H])=C9[H])C([H])=C8[H])<-[N]([H])([H])C([H])"
                "([H])C([H])([H])[N]->7([H])[H])C([H])=C6[H])C([H])=C5[H])<-"
                "[N]([H])([H])C([H])([H])C([H])([H])[N]->4([H])[H])C([H])=C3"
                "[H]"
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                stk.cage.M4L4Square(
                    corners=_get_palladium_cispbi_sqpl(),
                    linkers=get_other_linker(),
                    reaction_factory=stk.DativeReactionFactory(
                        stk.GenericReactionFactory(
                            bond_orders={
                                frozenset(
                                    {
                                        stk.GenericFunctionalGroup,
                                        stk.GenericFunctionalGroup,
                                    }
                                ): 9,
                            }
                        )
                    ),
                    scale_multiplier=2.0,
                )
            ),
            smiles=(
                "[H]C1=C2C3=C([H])C([H])=[N](->[Pd+2]4(<-[N]5=C([H])C([H])=C("
                "C6=C([H])C([H])=[N](->[Pd+2]7(<-[N]8=C([H])C([H])=C(C9=C([H]"
                ")C([H])=[N](->[Pd+2]%10(<-[N]%11=C([H])C([H])=C(C%12=C([H])C"
                "([H])=[N](->[Pd+2]%13(<-[N](=C1[H])C([H])=C2[H])<-[N]([H])(["
                "H])C([H])([H])C([H])([H])[N]->%13([H])[H])C([H])=C%12[H])C(["
                "H])=C%11[H])<-[N]([H])([H])C([H])([H])C([H])([H])[N]->%10([H"
                "])[H])C([H])=C9[H])C([H])=C8[H])<-[N]([H])([H])C([H])([H])C("
                "[H])([H])[N]->7([H])[H])C([H])=C6[H])C([H])=C5[H])<-[N]([H])"
                "([H])C([H])([H])C([H])([H])[N]->4([H])[H])C([H])=C3[H]"
            ),
            name=name,
        ),
    ),
)
def metal_cage_m4l4_square(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



import numpy as np
import pytest
import stk

from .case_data import CaseData

bb1 = stk.BuildingBlock("[C+2][N+]Br", [stk.BromoFactory()])
canonical_bb1 = bb1.with_canonical_atom_ordering()

bb2 = stk.BuildingBlock("IS[O+]", [stk.IodoFactory()])
canonical_bb2 = bb2.with_canonical_atom_ordering()


@pytest.fixture(
    params=(
        lambda: CaseData(
            molecule=stk.BuildingBlock(
                smiles="Br[C+2][N+]Cl",
                functional_groups=[stk.BromoFactory()],
                placer_ids=(0,),
            ),
            result=stk.BuildingBlock.init(
                atoms=(
                    stk.Cl(0),
                    stk.Br(1),
                    stk.C(2, 2),
                    stk.N(3, 1),
                ),
                bonds=(
                    stk.Bond(stk.Cl(0), stk.N(3, 1), 1),
                    stk.Bond(stk.Br(1), stk.C(2, 2), 1),
                    stk.Bond(stk.C(2, 2), stk.N(3, 1), 1),
                ),
                position_matrix=np.array([]),
                functional_groups=(
                    stk.Bromo(
                        bromine=stk.Br(1),
                        atom=stk.C(2, 2),
                        bonders=(stk.C(2, 2),),
                        deleters=(stk.Br(1),),
                    ),
                ),
                placer_ids=(1,),
            ),
        ),
        lambda: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(bb1, bb2),
                    repeating_unit="AB",
                    num_repeating_units=1,
                ),
            ),
            result=stk.ConstructedMolecule.init(
                atoms=(
                    stk.C(0, 2),
                    stk.O(1, 1),
                    stk.N(2, 1),
                    stk.S(3),
                ),
                bonds=(
                    stk.Bond(stk.C(0, 2), stk.N(2, 1), 1),
                    stk.Bond(stk.O(1, 1), stk.S(3), 1),
                    stk.Bond(stk.N(2, 1), stk.S(3), 1),
                ),
                position_matrix=np.array([]),
                atom_infos=(
                    stk.AtomInfo(
                        atom=stk.C(0, 2),
                        building_block_atom=stk.C(0, 2),
                        building_block=canonical_bb1,
                        building_block_id=0,
                    ),
                    stk.AtomInfo(
                        atom=stk.O(1, 1),
                        building_block_atom=stk.O(0, 1),
                        building_block=canonical_bb2,
                        building_block_id=1,
                    ),
                    stk.AtomInfo(
                        atom=stk.N(2, 1),
                        building_block_atom=stk.N(2, 1),
                        building_block=canonical_bb1,
                        building_block_id=0,
                    ),
                    stk.AtomInfo(
                        atom=stk.S(3),
                        building_block_atom=stk.S(2),
                        building_block=canonical_bb2,
                        building_block_id=1,
                    ),
                ),
                bond_infos=(
                    stk.BondInfo(
                        bond=stk.Bond(stk.C(0, 2), stk.N(2, 1), 1),
                        building_block=canonical_bb1,
                        building_block_id=0,
                    ),
                    stk.BondInfo(
                        bond=stk.Bond(stk.O(1, 1), stk.S(3), 1),
                        building_block=canonical_bb2,
                        building_block_id=1,
                    ),
                    stk.BondInfo(
                        bond=stk.Bond(stk.N(2, 1), stk.S(3), 1),
                        building_block=None,
                        building_block_id=None,
                    ),
                ),
                num_building_blocks={
                    canonical_bb1: 1,
                    canonical_bb2: 1,
                },
            ),
        ),
    ),
)
def case_data(request):
    return request.param()



import typing
from collections.abc import Iterable, Iterator, Sequence

import numpy as np

from stk._internal.ea.crossover.molecule_crosser import MoleculeCrosser
from stk._internal.ea.crossover.record import CrossoverRecord

T = typing.TypeVar("T")


class RandomCrosser(typing.Generic[T]):
    """
    Use some other crosser at random.

    Examples:

        *Use One of Several Crossers at Random*

        .. testcode:: use-one-of-several-crossers-at-random

            import stk

            def get_functional_group_type(building_block):
                fg, = building_block.get_functional_groups(0)
                return type(fg)

            def get_num_functional_groups(building_block):
               return building_block.get_num_functional_groups()

            crosser = stk.RandomCrosser(
                crossers=(
                    stk.GeneticRecombination(get_functional_group_type),
                    stk.GeneticRecombination(get_num_functional_groups),
                ),
                random_seed=12,
            )

            record1 = stk.MoleculeRecord(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='O=CC(C=O)CC=O',
                            functional_groups=[stk.AldehydeFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles='NCCN',
                            functional_groups=[stk.PrimaryAminoFactory()],
                        ),
                    ),
                ),
            )
            record2 = stk.MoleculeRecord(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='O=CNC(C=O)CC=O',
                            functional_groups=[stk.AldehydeFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles='NCNCN',
                            functional_groups=[stk.PrimaryAminoFactory()],
                        ),
                    ),
                ),
            )

            # Use one of the component crossers at random.
            crossover_records1 = tuple(crosser.cross((record1, record2)))
            # A different crosser may get selected at random the second,
            # third, etc, time.
            crossover_records2 = tuple(crosser.cross((record1, record2)))

        .. testcode:: use-one-of-several-crossers-at-random
            :hide:

            _expected_results = (
                stk.ConstructedMolecule(
                    topology_graph=stk.cage.FourPlusSix(
                        building_blocks=(
                            stk.BuildingBlock(
                                smiles='O=CNC(C=O)CC=O',
                                functional_groups=[stk.AldehydeFactory()],
                            ),
                            stk.BuildingBlock(
                                smiles='NCNCN',
                                functional_groups=[
                                    stk.PrimaryAminoFactory(),
                                ],
                            ),
                        ),
                    ),
                ),
                stk.ConstructedMolecule(
                    topology_graph=stk.cage.FourPlusSix(
                        building_blocks=(
                            stk.BuildingBlock(
                                smiles='O=CC(C=O)CC=O',
                                functional_groups=[stk.AldehydeFactory()],
                            ),
                            stk.BuildingBlock(
                                smiles='NCCN',
                                functional_groups=[
                                    stk.PrimaryAminoFactory(),
                                ],
                            ),
                        ),
                    ),
                ),
                stk.ConstructedMolecule(
                    topology_graph=stk.cage.FourPlusSix(
                        building_blocks=(
                            stk.BuildingBlock(
                                smiles='O=CNC(C=O)CC=O',
                                functional_groups=[stk.AldehydeFactory()],
                            ),
                            stk.BuildingBlock(
                                smiles='NCCN',
                                functional_groups=[
                                    stk.PrimaryAminoFactory(),
                                ],
                            ),
                        ),
                    ),
                ),
                stk.ConstructedMolecule(
                    topology_graph=stk.cage.FourPlusSix(
                        building_blocks=(
                            stk.BuildingBlock(
                                smiles='O=CC(C=O)CC=O',
                                functional_groups=[stk.AldehydeFactory()],
                            ),
                            stk.BuildingBlock(
                                smiles='NCNCN',
                                functional_groups=[
                                    stk.PrimaryAminoFactory(),
                                ],
                            ),
                        ),
                    ),
                ),
            )

            def _get_smiles(item):
                if isinstance(item, stk.ConstructedMolecule):
                   return stk.Smiles().get_key(item)
                return stk.Smiles().get_key(
                    molecule=item.get_molecule_record().get_molecule(),
                 )

            _expected_smiles = set(map(_get_smiles, _expected_results))

            _crossover_smiles1 = set(map(_get_smiles, crossover_records1))
            _crossover_smiles2 = set(map(_get_smiles, crossover_records2))
            assert _expected_smiles == _crossover_smiles1
            assert _expected_smiles == _crossover_smiles2

    """

    def __init__(
        self,
        crossers: Iterable[MoleculeCrosser[T]],
        weights: Iterable[float] | None = None,
        random_seed: int | np.random.Generator | None = None,
    ) -> None:
        """
        Parameters:

            crossers (list[MoleculeCrosser[T]]):
                A selection of crossers, each time :meth:`.cross` is called,
                one will be selected at random to perform crossover.

            weights (list[float] | None):
                For each crosser, the probability that it will be chosen
                whenever :meth:`.cross` is called.
                If ``None`` all `crossers` will have equal chance of being
                selected.

            random_seed:
                The random seed to use.
        """

        if weights is not None:
            weights = tuple(weights)

        if random_seed is None or isinstance(random_seed, int):
            random_seed = np.random.default_rng(random_seed)

        self._crossers = tuple(crossers)
        self._weights = weights
        self._generator = random_seed

    def cross(self, records: Sequence[T]) -> Iterator[CrossoverRecord[T]]:
        """
        Cross `records`.

        Parameters:
            molecules (list[T]):
                The molecules on which a crossover operation is
                performed.

        Yields:
            A record of a crossover operation.

        """

        crosser = self._generator.choice(
            a=self._crossers,  # type: ignore
            p=self._weights,
        )
        return crosser.cross(records)



import typing

import numpy as np

from stk._internal.key_makers.inchi_key import InchiKey
from stk._internal.key_makers.molecule import MoleculeKeyMaker
from stk._internal.molecule import Molecule
from stk._internal.utilities.utilities import OneOrMany

from .utilities import (
    AtomJson,
    BondJson,
    atom_to_json,
    bond_to_json,
    to_atom,
    to_bond,
)


class _MolecularGraphJson(typing.TypedDict):
    a: tuple[AtomJson, ...]
    b: tuple[BondJson, ...]


class _PositionMatrixJson(typing.TypedDict):
    m: list[list[float]]


class _MoleculeJson(typing.TypedDict):
    molecule: _MolecularGraphJson
    matrix: _PositionMatrixJson


class MoleculeJsonizer:
    """
    Abstract base class for creating JSONs of molecules.

    See Also:

        :class:`.ConstructedMoleculeJsonizer`

    Notes:

        You might notice that the public methods of this abstract base
        class are implemented. These are just default implementations,
        which can be safely ignored or overridden, when implementing
        subclasses. However, the default implementation can be used
        directly, if it suits your needs.

    Examples:

        *Converting a Molecule to JSON*

        You want to create a JSON representation of a molecule

        .. testcode:: converting-a-molecule-to-json

            import stk

            jsonizer = stk.MoleculeJsonizer()
            json = jsonizer.to_json(stk.BuildingBlock('NCCN'))

        *Adding Additional Molecular Keys*

        Apart from atoms, bonds and the position matrix, the JSON
        representation holds additional fields, one for each
        :class:`.MoleculeKeyMaker` provided to the initializer

        .. testcode:: adding-additional-molecular-keys

            import stk

            jsonizer = stk.MoleculeJsonizer(
                key_makers=(
                    stk.Inchi(),
                    stk.InchiKey(),
                ),
            )
            json = jsonizer.to_json(stk.BuildingBlock('NCCN'))

        In this case, ``json`` will look something like

        .. code-block:: python

            {
                # A tuple of JSON atom representations.
                'atoms': (...),

                # A tuple of JSON bond representations.
                'bonds': (...),

                'InChI': 'The InChI of the molecule',
                'InChIKey': 'The InChIKey of the molecule',
            }

        For every :class:`.MoleculeKeyMaker` provided to `key_makers`,
        a new key will be added to the JSON representation, with its
        name given by :meth:`.MoleculeKeyMaker.get_key_name` and the
        value given by :meth:`.MoleculeKeyMaker.get_key`.

    """

    def __init__(
        self,
        key_makers: OneOrMany[MoleculeKeyMaker] = (InchiKey(),),
    ) -> None:
        """
        Initialize a :class:`.MoleculeJsonizer` instance.

        Parameters:

            key_makers:
                Used to make the keys of molecules, which should be
                included in their JSON representations. Keys allow
                molecular data to reference itself when split across
                multiple JSONs.

        """

        if isinstance(key_makers, MoleculeKeyMaker):
            key_makers = (key_makers,)

        self._key_makers = tuple(key_makers)

    def to_json(
        self,
        molecule: Molecule,
    ) -> _MoleculeJson:
        """
        Return a JSON representation of `molecule`.

        Parameters:

            molecule:
                The molecule to serialize.

        Returns:

            A JSON representation of `molecule`.

        """

        json: _MolecularGraphJson = {
            "a": tuple(map(atom_to_json, molecule.get_atoms())),
            "b": tuple(map(bond_to_json, molecule.get_bonds())),
        }
        position_matrix: _PositionMatrixJson = {
            "m": molecule.get_position_matrix().tolist(),
        }
        for key_maker in self._key_makers:
            key_name = key_maker.get_key_name()
            key = key_maker.get_key(molecule)
            # TypedDict does not allow keys which are not listed in the
            # class definition. However, because we wish to create
            # indices on keys created by KeyMakers we need to add
            # these additional key-value pairs.
            json[key_name] = key  # type: ignore
            position_matrix[key_name] = key  # type: ignore
        return {
            "molecule": json,
            "matrix": position_matrix,
        }

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._key_makers!r})"


class MoleculeDejsonizer:
    """
    Abstract base class for creating molecules from JSONs.

    See Also
    --------
    :class:`.ConstructedMoleculeDejsonizer`

    Notes
    -----
    You might notice that the public methods of this abstract base
    class are implemented. These are just default implementations,
    which can be safely ignored or overridden, when implementing
    subclasses. However, the default implementation can be used
    directly, if it suits your needs.

    """

    def from_json(self, json):
        """
        Get a :class:`.Molecule` from a JSON.

        Parameters
        ----------
        json : :class:`dict`
            A JSON representation of a molecule.

        Returns
        -------
        :class:`.Molecule`
            The molecule held in `json`.

        """

        atoms = tuple(
            to_atom(atom_id, atom_json)
            for atom_id, atom_json in enumerate(json["molecule"]["a"])
        )
        return Molecule(
            atoms=atoms,
            bonds=tuple(
                to_bond(atoms, bond_json)
                for bond_json in json["molecule"]["b"]
            ),
            position_matrix=np.array(json["matrix"]["m"]),
        )

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f"{self.__class__.__name__}()"



"""
Octahedral Delta
================

"""

from stk._internal.topology_graphs.edge import Edge

from .metal_complex import MetalComplex
from .vertices import BiDentateLigandVertex, MetalVertex


class OctahedralDelta(MetalComplex):
    """
    Represents a metal complex topology graph.

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='[Fe+2]',
            functional_groups=(
                stk.SingleAtom(stk.Fe(0, charge=2))
                for i in range(6)
            ),
            position_matrix=[[0, 0, 0]],
        )
        bb2 = stk.BuildingBlock(
            smiles=(
                'C1=CC=NC(C2=CC=CC(C3=C'
                'C=CC=N3)=C2)=C1'
            ),
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )
        complex = stk.ConstructedMolecule(
            topology_graph=stk.metal_complex.OctahedralDelta(
                metals=bb1,
                ligands=bb2,
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    complex.get_atoms(),
                    complex.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in complex.get_bonds()
            ),
        )

    Metal building blocks with at least six functional groups are
    required for this topology.

    Ligand building blocks with two functional groups are required for
    this topology graph.

    When using a :class:`dict` for initialization, a
    :class:`.BuildingBlock` needs to be assigned to each of the
    following numbers:

        | metals: (0, )
        | ligands: (0, 1, 2)

    See :class:`.MetalComplex` for more details and examples.

    """

    _metal_vertex_prototypes = (MetalVertex(0, (0, 0, 0)),)
    _ligand_vertex_prototypes = (
        BiDentateLigandVertex(1, (2.5, 2.5, 0)),
        BiDentateLigandVertex(2, (0, -2.5, 2.5)),
        BiDentateLigandVertex(3, (-2.5, 0, -2.5)),
    )

    # The ordering here matters for the stereochemistry.
    # The first edge to appear between two vertices determines the
    # directionality of the binding ligand.
    _edge_prototypes = (
        Edge(
            id=0,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[0],
            position=(2.5, 0, 0),
        ),
        Edge(
            id=1,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[0],
            position=(0, 2.5, 0),
        ),
        Edge(
            id=4,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[1],
            position=(0, -2.5, 0),
        ),
        Edge(
            id=2,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[1],
            position=(0, 0, 2.5),
        ),
        Edge(
            id=5,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[2],
            position=(0, 0, -2.5),
        ),
        Edge(
            id=3,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[2],
            position=(-2.5, 0, 0),
        ),
    )



import itertools as it


def test_get_atoms(case_data):
    """
    Test :meth:`.FunctionalGroup.get_atoms`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        The test case. Holds the functional group to test and the
        correct atoms.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_atoms(
        functional_group=case_data.functional_group,
        atoms=case_data.atoms,
    )


def _test_get_atoms(functional_group, atoms):
    """
    Test :meth:`.FunctionalGroup.get_atoms`.

    Parameters
    ----------
    functional_group : :class:`.FunctionalGroup`
        The functional group to test.

    atoms : :class:`tuple` of :class:`.Atom`
        The correct atoms.

    Returns
    -------
    None : :class:`NoneType`

    """

    for atom1, atom2 in it.zip_longest(functional_group.get_atoms(), atoms):
        assert atom1 is atom2



import typing
from dataclasses import dataclass

import stk


@dataclass(frozen=True, slots=True)
class CaseData:
    fitness_calculator: stk.FitnessCalculator
    record: stk.MoleculeRecord
    fitness_value: typing.Any



import pytest

import stk

from .case_data import CaseData


@pytest.fixture(
    params=(
        lambda name: CaseData(
            molecules=[
                stk.BuildingBlock(smiles="CC"),
                stk.BuildingBlock(smiles="CNC"),
                stk.BuildingBlock(smiles="CNNC"),
                stk.BuildingBlock(smiles="CNC1CCC1C"),
            ],
            property_dicts=[
                {"1": 2, "3": "astr"},
                {"1": 2, "3": "bstr"},
                {"1": 2, "3": "cstr"},
                {"1": 3, "3": "cstr", "2": {"h": "w"}},
            ],
            expected_count=4,
            name=name,
        ),
    )
)
def molecule(request: pytest.FixtureRequest) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "stk"
copyright = "2023, Lukas Turcani"
author = "Lukas Turcani"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.doctest",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
    "moldoc",
]

autosummary_imported_members = True

autodoc_typehints = "description"
autodoc_member_order = "groupwise"
autodoc_type_aliases = {
    "NumBuildingBlocks": "dict[Molecule, int]",
    "IncludedBatches": "set[BatchKey] | None",
    "ExcludedBatches": "set[BatchKey] | None",
}
autoclass_content = "both"

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pandas": ("https://pandas.pydata.org/docs/", None),
}

templates_path = ["_templates"]
exclude_patterns: list[str] = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



import itertools as it

import numpy as np
import pytest
import stk

from ..case_data import CaseData
from .utilities import MockConstructionState, MockEdge


@pytest.fixture
def functional_group2_2(functional_group2):
    """
    A :class:`.GenericFunctionalGroup` with 2 bonder atoms.

    """

    return functional_group2.clone()


@pytest.fixture
def two_two_reaction(
    periodicity,
    functional_group2,
    functional_group2_2,
    bond_order,
):
    bond_order_key = frozenset(
        {
            type(functional_group2),
            type(functional_group2_2),
        }
    )
    position_matrix = get_position_matrix(
        functional_group1=functional_group2,
        functional_group2=functional_group2_2,
    )
    edge = MockEdge(0, periodicity)
    return CaseData(
        factory=stk.GenericReactionFactory(
            bond_orders={
                bond_order_key: bond_order,
            },
        ),
        construction_state=MockConstructionState(
            edges=(edge,),
            edge_functional_groups={
                0: (
                    functional_group2,
                    functional_group2_2,
                ),
            },
            position_matrix=position_matrix,
        ),
        edge_group=stk.EdgeGroup(
            edges=(edge,),
        ),
        reaction_result=stk.ReactionResult(
            new_atoms=(),
            new_bonds=get_new_bonds(
                functional_group1=functional_group2,
                functional_group2=functional_group2_2,
                order=bond_order,
                periodicity=periodicity,
            ),
            deleted_atoms=it.chain(
                functional_group2.get_deleters(),
                functional_group2_2.get_deleters(),
            ),
            deleted_bonds=(),
        ),
    )


def get_new_bonds(
    functional_group1,
    functional_group2,
    order,
    periodicity,
):
    bonder1, bonder2 = functional_group1.get_bonders()
    bonder3, bonder4 = functional_group2.get_bonders()
    yield stk.Bond(
        atom1=bonder1,
        atom2=bonder3,
        order=order,
        periodicity=periodicity,
    )
    yield stk.Bond(
        atom1=bonder2,
        atom2=bonder4,
        order=order,
        periodicity=periodicity,
    )


def get_position_matrix(functional_group1, functional_group2):
    size = (
        max(
            it.chain(
                functional_group1.get_bonder_ids(),
                functional_group2.get_bonder_ids(),
            )
        )
        + 1
    )
    position_matrix = np.zeros((size, 3))
    bonder1, bonder2 = functional_group1.get_bonder_ids()
    bonder3, bonder4 = functional_group2.get_bonder_ids()

    # Place the first bonder of each functional group close.
    position_matrix[bonder1] = [100, 100, 100]
    position_matrix[bonder3] = [99, 100, 100]
    return position_matrix
